# Copyright (C) 2023 Caqui - All Rights Reserved
# You may use, distribute and modify this code under the
# terms of the MIT license.
# Visit: https://github.com/douglasdcm/caqui

from typing import TYPE_CHECKING, Union

from caqui.webdriver.alert import Alert
from caqui.webdriver.element import Element
from caqui.webdriver.engine import asynchronous, synchronous

if TYPE_CHECKING:
    from caqui.webdriver.drivers import AsyncDriver


class SwitchToW3C:
    def __init__(self, driver: "AsyncDriver") -> None:
        self._driver = driver
        self._iframe: Union[str] = ""
        self._window_handle: Union[str] = ""
        self._session_http = driver.session_http

    @property
    def active_element(self) -> "Element":
        """Returns the active element"""
        element = synchronous.get_active_element(self._driver.server_url, self._driver.session)
        return Element(element, self._driver)

    @property
    def alert(self) -> "Alert":
        """Returns the `Alert` object"""
        return Alert(self._driver)

    async def new_window(self, window_type) -> str:
        """Opens a new window"""
        self._window_handle = await asynchronous.new_window(
            self._driver.server_url,
            self._driver.session,
            window_type,
            session_http=self._session_http,
        )
        self._window_handle = await asynchronous.switch_to_window(
            self._driver.server_url,
            self._driver.session,
            self._window_handle,
            session_http=self._session_http,
        )
        return self._window_handle

    async def window(self, window_handle: str) -> str:
        """Switchs to window `window_handle`"""
        self._window_handle = await asynchronous.switch_to_window(
            self._driver.server_url,
            self._driver.session,
            window_handle,
            session_http=self._session_http,
        )
        return self._window_handle

    async def frame(self, iframe: Union[str, Element]) -> None:
        """Switches to frame `iframe`"""
        self._iframe = str(iframe)
        await asynchronous.switch_to_frame(
            self._driver.server_url,
            self._driver.session,
            self._iframe,
            session_http=self._session_http,
        )

    async def default_content(self) -> None:
        """Switches to parent frame of 'element_frame'"""
        await asynchronous.switch_to_parent_frame(
            self._driver.server_url,
            self._driver.session,
            self._iframe,
            session_http=self._session_http,
        )


class SwitchToJsonWire(SwitchToW3C):
    def __init__(self, driver) -> None:
        super().__init__(driver)
        self._driver = driver
        self._iframe: Union[str] = ""
        self._window_handle: Union[str] = ""
        self._session_http = driver.session_http

    @property
    def active_element(self):
        """Returns the active element"""
        element = synchronous.get_active_element_jsonwire(
            self._driver.server_url, self._driver.session
        )
        return Element(element, self._driver)

    async def new_window(self, window_type):
        """Opens a new window"""
        self._window_handle = await asynchronous.new_window(
            self._driver.server_url,
            self._driver.session,
            window_type,
            session_http=self._session_http,
        )
        self._window_handle = await asynchronous.switch_to_window_jsonwire(
            self._driver.server_url,
            self._driver.session,
            self._window_handle,
            session_http=self._session_http,
        )
        return self._window_handle

    async def window(self, window_handle):
        """Switchs to window `window_handle`"""
        self._window_handle = await asynchronous.switch_to_window_jsonwire(
            self._driver.server_url,
            self._driver.session,
            window_handle,
            session_http=self._session_http,
        )
        return self._window_handle

    async def default_content(self):
        """Switches to parent frame of 'element_frame'"""
        return await asynchronous.switch_to_parent_frame_jsonwire(
            self._driver.server_url,
            self._driver.session,
            self._iframe,
            session_http=self._session_http,
        )

    async def frame(self, iframe):
        """Switches to frame `iframe`"""
        self._iframe = str(iframe)
        return await asynchronous.switch_to_frame_jsonwire(
            self._driver.server_url,
            self._driver.session,
            self._iframe,
            session_http=self._session_http,
        )



# Copyright (C) 2023 Caqui - All Rights Reserved
# You may use, distribute and modify this code under the
# terms of the MIT license.
# Visit: https://github.com/douglasdcm/caqui

from caqui.cdp.connection import SyncCDPConnection  # noqa F401



"""
CDP Connection Module

This module provides I/O and multiplexing support for Chrome DevTools Protocol.
It handles WebSocket connections, JSON-RPC message framing, command multiplexing,
and event dispatching.
"""

from __future__ import annotations

import asyncio
import json
import logging
import typing
from dataclasses import dataclass, field

try:
    import websockets
    from websockets.client import WebSocketClientProtocol
    WEBSOCKETS_AVAILABLE = True
except ImportError:
    WEBSOCKETS_AVAILABLE = False
    WebSocketClientProtocol = typing.Any  # type: ignore

from caqui._vendor.chrome_devtools_protocol.cdp.util import T_JSON_DICT, parse_json_event

logger = logging.getLogger(__name__)


class CDPError(Exception):
    """Base exception for CDP errors."""
    pass


class CDPConnectionError(CDPError):
    """Raised when there's a connection error."""
    pass


class CDPCommandError(CDPError):
    """Raised when a command returns an error."""
    
    def __init__(self, code: int, message: str, data: typing.Optional[typing.Any] = None):
        self.code = code
        self.message = message
        self.data = data
        super().__init__(f"CDP Command Error {code}: {message}")


@dataclass
class PendingCommand:
    """Represents a command waiting for a response."""
    future: asyncio.Future
    method: str
    params: T_JSON_DICT


class CDPConnection:
    """
    Manages a WebSocket connection to Chrome DevTools Protocol.
    
    This class handles:
    - WebSocket connection management
    - JSON-RPC message framing (request ID assignment)
    - Command multiplexing (tracking multiple concurrent commands)
    - Event dispatching
    - Error handling
    
    Example:
        async with CDPConnection("ws://localhost:9222/devtools/page/...") as conn:
            # Send a command
            result = await conn.execute(some_command())
            
            # Listen for events
            async for event in conn.listen():
                print(event)
    """
    
    def __init__(self, url: str, timeout: float = 30.0):
        """
        Initialize a CDP connection.
        
        Args:
            url: WebSocket URL for the CDP endpoint
            timeout: Default timeout for commands in seconds
        """
        if not WEBSOCKETS_AVAILABLE:
            raise ImportError(
                "websockets library is required for CDPConnection. "
                "Install it with: pip install websockets"
            )
        
        self.url = url
        self.timeout = timeout
        self._ws: typing.Optional[WebSocketClientProtocol] = None
        self._next_command_id = 1
        self._pending_commands: typing.Dict[int, PendingCommand] = {}
        self._event_queue: asyncio.Queue = asyncio.Queue()
        self._recv_task: typing.Optional[asyncio.Task] = None
        self._closed = False
    
    async def connect(self) -> None:
        """Establish the WebSocket connection."""
        if self._ws is not None:
            raise CDPConnectionError("Already connected")
        
        try:
            self._ws = await websockets.connect(self.url)  # type: ignore
            self._recv_task = asyncio.create_task(self._receive_loop())
            logger.info(f"Connected to {self.url}")
        except Exception as e:
            raise CDPConnectionError(f"Failed to connect to {self.url}: {e}")
    
    async def close(self) -> None:
        """Close the WebSocket connection."""
        if self._closed:
            return
        
        self._closed = True
        
        # Cancel the receive task
        if self._recv_task:
            self._recv_task.cancel()
            try:
                await self._recv_task
            except asyncio.CancelledError:
                pass
        
        # Cancel all pending commands
        for cmd_id, pending in self._pending_commands.items():
            if not pending.future.done():
                pending.future.cancel()
        self._pending_commands.clear()
        
        # Close the WebSocket
        if self._ws:
            await self._ws.close()
            self._ws = None
        
        logger.info("Connection closed")
    
    async def __aenter__(self) -> CDPConnection:
        """Async context manager entry."""
        await self.connect()
        return self
    
    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """Async context manager exit."""
        await self.close()
    
    async def _receive_loop(self) -> None:
        """
        Main receive loop that processes incoming WebSocket messages.
        
        This loop:
        - Receives messages from the WebSocket
        - Parses JSON-RPC responses and matches them to pending commands
        - Dispatches events to the event queue
        """
        try:
            while not self._closed and self._ws:
                try:
                    message = await self._ws.recv()
                    data = json.loads(message)
                    
                    if 'id' in data:
                        # This is a command response
                        await self._handle_response(data)
                    elif 'method' in data:
                        # This is an event
                        await self._handle_event(data)
                    else:
                        logger.warning(f"Received unexpected message: {data}")
                
                except json.JSONDecodeError as e:
                    logger.error(f"Failed to decode JSON: {e}")
                except Exception as e:
                    logger.error(f"Error in receive loop: {e}")
                    if not self._closed:
                        raise
        except asyncio.CancelledError:
            logger.debug("Receive loop cancelled")
        except Exception as e:
            logger.error(f"Fatal error in receive loop: {e}")
            # Cancel all pending commands with this error
            for pending in self._pending_commands.values():
                if not pending.future.done():
                    pending.future.set_exception(CDPConnectionError(f"Connection error: {e}"))
    
    async def _handle_response(self, data: T_JSON_DICT) -> None:
        """Handle a command response."""
        cmd_id = data['id']
        
        if cmd_id not in self._pending_commands:
            logger.warning(f"Received response for unknown command ID {cmd_id}")
            return
        
        pending = self._pending_commands.pop(cmd_id)
        
        if 'error' in data:
            error = data['error']
            exc = CDPCommandError(
                code=error.get('code', -1),
                message=error.get('message', 'Unknown error'),
                data=error.get('data')
            )
            pending.future.set_exception(exc)
        else:
            result = data.get('result', {})
            pending.future.set_result(result)
    
    async def _handle_event(self, data: T_JSON_DICT) -> None:
        """Handle an event notification."""
        try:
            event = parse_json_event(data)
            await self._event_queue.put(event)
        except Exception as e:
            logger.error(f"Failed to parse event: {e}")
    
    async def execute(
        self,
        cmd: typing.Generator[T_JSON_DICT, T_JSON_DICT, typing.Any],
        timeout: typing.Optional[float] = None
    ) -> typing.Any:
        """
        Execute a CDP command.
        
        This method:
        - Assigns a unique ID to the command
        - Sends it over the WebSocket
        - Waits for the response (with multiplexing support)
        - Returns the parsed result
        
        Args:
            cmd: A CDP command generator (from any CDP domain module)
            timeout: Optional timeout override for this command
        
        Returns:
            The command result (type depends on the command)
        
        Raises:
            CDPCommandError: If the command returns an error
            asyncio.TimeoutError: If the command times out
            CDPConnectionError: If there's a connection error
        
        Example:
            from caqui._vendor.chrome_devtools_protocol.cdp import page
            result = await conn.execute(page.navigate(url="https://example.com"))
        """
        if self._ws is None:
            raise CDPConnectionError("Not connected")
        
        if self._closed:
            raise CDPConnectionError("Connection closed")
        
        # Get the command request from the generator
        request = cmd.send(None)  # type: ignore[arg-type]
        
        # Assign a unique ID
        cmd_id = self._next_command_id
        self._next_command_id += 1
        request['id'] = cmd_id
        
        # Create a future to track this command
        future: asyncio.Future = asyncio.Future()
        self._pending_commands[cmd_id] = PendingCommand(
            future=future,
            method=request['method'],
            params=request.get('params', {})
        )
        
        try:
            # Send the command
            await self._ws.send(json.dumps(request))
            logger.debug(f"Sent command {cmd_id}: {request['method']}")
            
            # Wait for the response
            timeout_val = timeout if timeout is not None else self.timeout
            result = await asyncio.wait_for(future, timeout=timeout_val)
            
            # Send the result back to the generator
            try:
                cmd.send(result)
            except StopIteration as e:
                return e.value
            
            raise CDPError("Command generator did not stop")
            
        except asyncio.TimeoutError:
            # Clean up the pending command
            self._pending_commands.pop(cmd_id, None)
            raise asyncio.TimeoutError(f"Command {request['method']} timed out")
        except Exception:
            # Clean up the pending command on error
            self._pending_commands.pop(cmd_id, None)
            raise
    
    async def listen(self) -> typing.AsyncIterator[typing.Any]:
        """
        Listen for events from the browser.
        
        This is an async iterator that yields CDP events as they arrive.
        
        Yields:
            CDP event objects (type depends on the event)
        
        Example:
            async for event in conn.listen():
                if isinstance(event, page.LoadEventFired):
                    print("Page loaded!")
        """
        while not self._closed:
            try:
                event = await asyncio.wait_for(self._event_queue.get(), timeout=1.0)
                yield event
            except asyncio.TimeoutError:
                # Check if connection is still alive
                if self._closed:
                    break
                continue
    
    def get_event_nowait(self) -> typing.Optional[typing.Any]:
        """
        Get an event from the queue without waiting.
        
        Returns:
            A CDP event object, or None if no events are available
        """
        try:
            return self._event_queue.get_nowait()
        except asyncio.QueueEmpty:
            return None
    
    @property
    def is_connected(self) -> bool:
        """Check if the connection is open."""
        return self._ws is not None and not self._closed
    
    @property
    def pending_command_count(self) -> int:
        """Get the number of pending commands (for debugging/monitoring)."""
        return len(self._pending_commands)



import datetime
from asyncio import sleep as async_sleep
from time import sleep
from typing import Callable, Coroutine

from caqui.constants import TIMEOUT


class WebDriverWait:
    def __init__(self, driver, timeout=TIMEOUT, pooling_time=0.5):
        self._driver = driver
        self._timeout = timeout
        self._pooling_time = pooling_time

    def until(self, condition: Callable):
        """Waits a condition be true or raises a TimeoutError exception"""
        current_datetime = datetime.datetime.now()
        time_to_add = datetime.timedelta(seconds=self._timeout)
        new_datetime = current_datetime + time_to_add
        while datetime.datetime.now() < new_datetime:
            if condition():
                return
            sleep(self._pooling_time)
        raise TimeoutError()

    async def async_until(self, condition: Coroutine):
        """Waits a condition be true or raises a TimeoutError exception"""
        current_datetime = datetime.datetime.now()
        time_to_add = datetime.timedelta(seconds=self._timeout)
        new_datetime = current_datetime + time_to_add
        while datetime.datetime.now() < new_datetime:
            if await condition():
                return
            await async_sleep(self._pooling_time)
        raise TimeoutError()



# Copyright (C) 2023 Caqui - All Rights Reserved
# You may use, distribute and modify this code under the
# terms of the MIT license.
# Visit: https://github.com/douglasdcm/caqui

import asyncio
import datetime
import json
import urllib.request
from typing import Any, Dict, List, Optional
from urllib.parse import urlparse

from bs4 import BeautifulSoup

from caqui._vendor.chrome_devtools_protocol.cdp import (
    accessibility,
    browser,
    dom,
    emulation,
    input_,
    network,
    page,
    runtime,
    storage,
    target,
)
from caqui.cdp.asynchronous.connection import AsyncCDPConnection
from caqui.cdp.by import By
from caqui.constants import TIME_FORMAT_MICROSECONDS, TIMEOUT
from caqui.exceptions import WebDriverError
from caqui.helper import convert_locator_to_css_selector_or_xpath, save_picture


class GlobalValues:
    _doc: Optional[dom.Node] = None
    conn: AsyncCDPConnection = None

    async def get_document_node(refresh=False) -> dom.Node:
        if refresh:
            GlobalValues._doc = None
        if not GlobalValues.conn:
            raise WebDriverError("No connection")
        if not GlobalValues._doc:
            GlobalValues._doc = await GlobalValues.conn.execute(
                dom.get_document(depth=-1, pierce=True)
            )
        return GlobalValues._doc


async def _get_element_center(conn, element):
    box_model = await conn.execute(dom.get_box_model(node_id=element))
    content_box = box_model.content
    center_x = (content_box[0] + content_box[2]) / 2
    center_y = (content_box[1] + content_box[5]) / 2
    return center_x, center_y


async def _find_element_by_xpath(
    conn: AsyncCDPConnection,
    xpath: str,
    root_id: Optional[dom.NodeId] = None,
) -> dom.NodeId:
    if root_id:
        remote_root = await conn.execute(dom.resolve_node(node_id=root_id))
    else:
        node = await GlobalValues.get_document_node()
        remote_root = await conn.execute(dom.resolve_node(node_id=node.node_id))
    object_id = remote_root.object_id
    call_res = await conn.execute(
        runtime.call_function_on(
            object_id=object_id,
            function_declaration=f"""
            function() {{
                const doc = this.ownerDocument || this;
                const r = doc.evaluate(
                    `{xpath}`,
                    this,
                    null,
                    XPathResult.FIRST_ORDERED_NODE_TYPE,
                    null
                );
                return r.singleNodeValue;
            }}
            """,
            return_by_value=False,
        )
    )
    remote = call_res[0]
    if not remote.object_id:
        raise WebDriverError("XPath returned no element")
    node = await conn.execute(dom.request_node(object_id=remote.object_id))
    if node == 0:
        raise WebDriverError("Resolved object is not a DOM node")
    return node


async def _find_all_elements_by_xpath(
    conn: AsyncCDPConnection,
    xpath: str,
    root_id: Optional[dom.NodeId] = None,
) -> List[dom.NodeId]:
    if root_id:
        remote_root = await conn.execute(dom.resolve_node(node_id=root_id))
    else:
        node = await GlobalValues.get_document_node()
        remote_root = await conn.execute(dom.resolve_node(node_id=node.node_id))

    object_id = remote_root.object_id

    call_res = await conn.execute(
        runtime.call_function_on(
            object_id=object_id,
            function_declaration=f"""
            function () {{
                const doc = this.ownerDocument || this;
                const snapshot = doc.evaluate(
                    `{xpath}`,
                    this,
                    null,
                    XPathResult.ORDERED_NODE_SNAPSHOT_TYPE,
                    null
                );

                const results = [];
                for (let i = 0; i < snapshot.snapshotLength; i++) {{
                    results.push(snapshot.snapshotItem(i));
                }}
                return results;
            }}
            """,
            return_by_value=False,
        )
    )

    remote_array = call_res[0]

    if not remote_array.object_id:
        raise WebDriverError("XPath returned no elements")

    properties = await conn.execute(
        runtime.get_properties(object_id=remote_array.object_id, own_properties=True)
    )

    nodes: List[dom.NodeId] = []

    for prop in properties[0]:
        if not prop.name.isdigit():
            continue

        remote_node = prop.value
        if not remote_node or not remote_node.object_id:
            continue

        node_id = await conn.execute(dom.request_node(object_id=remote_node.object_id))

        if node_id != 0:
            nodes.append(node_id)

    if not nodes:
        raise WebDriverError("Resolved objects are not DOM nodes")

    return nodes


async def describe_node_id(conn: AsyncCDPConnection, node_id: dom.NodeId):
    if not node_id:
        raise WebDriverError(f"Invalid node id '{node_id}'")
    return await conn.execute(dom.describe_node(node_id))


async def _find_element_by_css_selector(
    conn: AsyncCDPConnection, root_id: dom.NodeId, depth: int, pierce: bool, selector: str
):
    if root_id:
        node = await conn.execute(dom.describe_node(root_id))
    else:
        node = await GlobalValues.get_document_node()
    node_id = await conn.execute(dom.query_selector(node_id=node.node_id, selector=selector))
    if not node_id:
        raise WebDriverError()
    return node_id


async def _find_element(
    conn: AsyncCDPConnection,
    locator_type,
    locator_value: str,
    root_id: dom.NodeId = None,
    depth=1,
    pierce=False,
):
    locator_type, selector = convert_locator_to_css_selector_or_xpath(locator_type, locator_value)
    if locator_type == By.XPATH:
        try:
            return await _find_element_by_xpath(conn, selector, root_id)
        except Exception:
            raise
    if locator_type == By.CSS_SELECTOR:
        return await _find_element_by_css_selector(conn, root_id, depth, pierce, selector)
    raise WebDriverError(f"Could not find element with selector: {locator_value}")


async def _find_all_elements(
    conn: AsyncCDPConnection,
    locator_type,
    locator_value: str,
    root_id: dom.NodeId = None,
):
    locator_type, locator_value = convert_locator_to_css_selector_or_xpath(
        locator_type, locator_value
    )
    if locator_type == By.XPATH:
        return await _find_all_elements_by_xpath(conn, locator_value, root_id)
    if locator_type == By.CSS_SELECTOR:
        if root_id:
            node = await conn.execute(dom.describe_node(root_id))
        else:
            node = await GlobalValues.get_document_node()
        node_id = await conn.execute(
            dom.query_selector_all(node_id=node.node_id, selector=locator_value)
        )
        if not node_id:
            raise WebDriverError()
        return node_id


async def _handle_alert(
    conn: AsyncCDPConnection, alert_element, timeout, function_callback, text=None
):
    click_task = asyncio.create_task(click(conn, alert_element))
    current_datetime = datetime.datetime.now()
    time_to_add = datetime.timedelta(seconds=timeout)
    new_datetime = current_datetime + time_to_add
    while 1 == 1:  # Used 1==1 and not True to deceive VSCode
        while datetime.datetime.now() < new_datetime:
            event = conn.get_event_nowait()
            if event is None:
                await asyncio.sleep(0)  # yield control
                continue
            if isinstance(event, page.JavascriptDialogOpening):
                await function_callback(conn, text)
                return event.message
        raise TimeoutError()
    await click_task


async def _send_test_to_prompt_alert(conn: AsyncCDPConnection, text):
    await conn.execute(page.handle_java_script_dialog(accept=True, prompt_text=text))


async def _accept_alert(conn: AsyncCDPConnection, text):
    await conn.execute(page.handle_java_script_dialog(accept=True))


async def _dismiss_alert(conn: AsyncCDPConnection, text):
    await conn.execute(page.handle_java_script_dialog(accept=False))


async def _set_screen(conn, screen_type: str = "fullscreen"):
    window_for_target = await conn.execute(browser.get_window_for_target())
    window_id = window_for_target[0]
    bounds = browser.Bounds.from_json({"windowState": screen_type})
    await conn.execute(browser.set_window_bounds(window_id=window_id, bounds=bounds))


async def get(conn: AsyncCDPConnection, page_url: str) -> None:
    """Does the same of 'go_to_page'. Added to be compatible with selenium method name'"""
    try:
        await conn.execute(page.enable())
        await conn.execute(page.navigate(url=page_url))
        await conn.execute(
            target.set_auto_attach(auto_attach=True, wait_for_debugger_on_start=False, flatten=True)
        )
        await _refresh_agents(conn)
    except Exception as e:
        raise WebDriverError(f"Failed to navigate to page '{page_url}'.") from e


async def _refresh_agents(conn: AsyncCDPConnection):
    await conn.execute(page.enable())
    await conn.execute(dom.enable())
    await conn.execute(runtime.enable())
    await conn.execute(network.enable())
    await conn.execute(accessibility.enable()),
    GlobalValues.conn = conn
    await GlobalValues.get_document_node(refresh=True)


async def go_to_page(conn, page_url: str) -> None:
    """Navigate to 'page_url'"""
    await get(conn, page_url)


async def find_element(
    conn: AsyncCDPConnection, locator_type: str, locator_value: str, root_id: dom.NodeId = None
) -> dom.NodeId:
    """Find an element by a 'selector', for example an 'xpath' like '//div[@id="example"]'

    `Attention`: it opens Alerts/Prompt elements automatically.
    Not applicable for elements in iframes.
    """
    try:
        return await _find_element(conn, locator_type, locator_value, root_id)
    except Exception as e:
        raise WebDriverError(f"Could not find element with selector: {locator_value}") from e


async def find_elements(conn: AsyncCDPConnection, locator_type, locator_value: str):
    """Find an element by a 'selector', for example an 'xpath' like '//div[@id="example"]'"""
    try:
        return await _find_all_elements(conn, locator_type, locator_value)
    except Exception as e:
        raise WebDriverError(f"Could not find element with selector: {locator_value}") from e


async def click(conn: AsyncCDPConnection, element):
    """Click on an element"""
    try:
        center_x, center_y = await _get_element_center(conn, element)
        commands = []
        mouse_button = input_.MouseButton("left")
        commands.append(
            input_.dispatch_mouse_event(
                "mousePressed", center_x, center_y, button=mouse_button, click_count=1
            )
        )
        commands.append(
            input_.dispatch_mouse_event(
                "mouseReleased", center_x, center_y, button=mouse_button, click_count=1
            )
        )
        for command in commands:
            await conn.execute(command)
    except Exception as e:
        raise WebDriverError("Failed to click on element.") from e


async def send_keys(conn: AsyncCDPConnection, element, text: str):
    """Send keys to an element"""
    try:
        await conn.execute(dom.focus(node_id=element))
        for char in text:
            await conn.execute(
                input_.dispatch_key_event(
                    "keyDown",
                    text=char,
                    unmodified_text=char,
                    windows_virtual_key_code=ord(char),
                    native_virtual_key_code=ord(char),
                )
            )
            await conn.execute(
                input_.dispatch_key_event(
                    "keyUp",
                    text=char,
                    unmodified_text=char,
                    windows_virtual_key_code=ord(char),
                    native_virtual_key_code=ord(char),
                )
            )
    except Exception as e:
        raise WebDriverError("Failed to send keys to element.") from e


async def get_text(conn: AsyncCDPConnection, element) -> str:
    """Get the text of an element"""
    try:
        result = await execute_script(conn, element, "value")
        if result and result[0].value:
            return result[0].value
        html_doc = await conn.execute(dom.get_outer_html(node_id=element))
        soup = BeautifulSoup(html_doc, "html.parser")
        return soup.get_text()
    except Exception as e:
        raise WebDriverError("Failed to get text from element.") from e


async def get_attribute(conn: AsyncCDPConnection, element, attribute: str) -> str:
    """Get the given HTML attribute of an element, for example, 'aria-valuenow'"""
    try:
        attributes = await conn.execute(dom.get_attributes(node_id=element))
        attr_dict = {attributes[i]: attributes[i + 1] for i in range(0, len(attributes), 2)}
        result = attr_dict.get(attribute)
        if result is None:
            raise WebDriverError("Failed to get value from element.")
        return result
    except Exception as e:
        raise WebDriverError("Failed to get value from element.") from e


async def get_title(conn: AsyncCDPConnection) -> str:
    """Get the page title"""
    try:
        target_info = await conn.execute(target.get_target_info())
        return target_info.title if target_info.type_ == "page" else ""
    except Exception:
        pass
    try:
        root_node = await GlobalValues.get_document_node()
        title_node_id = await conn.execute(
            dom.query_selector(node_id=root_node.node_id, selector="title")
        )
        result = await conn.execute(dom.get_outer_html(node_id=title_node_id))
        return BeautifulSoup(result, "html.parser").get_text()
    except Exception as e:
        raise WebDriverError("Failed to get page title.") from e


async def get_url(conn: AsyncCDPConnection) -> str:
    """Returns the URL from web page:"""
    try:
        frame_tree = await conn.execute(page.get_frame_tree())
        current_url = frame_tree.frame.url
        return current_url
    except Exception as e:
        raise WebDriverError("Failed to get page url.") from e


async def go_back(conn: AsyncCDPConnection):
    try:
        result = await conn.execute(page.get_navigation_history())
        current = result[0]
        if current <= 0:
            return
        entry = result[1][current - 1]
        await conn.execute(page.navigate_to_history_entry(entry.id_))
        await _refresh_agents(conn)
    except Exception as e:
        raise WebDriverError("Failed to go back to page.") from e


async def is_element_selected(conn: AsyncCDPConnection, element) -> bool:
    """Check if element is selected"""
    try:
        try:
            get_attribute_value = await get_attribute(conn, element, "checked")
            return get_attribute_value.lower() == ""
        except WebDriverError:
            pass
        try:
            get_attribute_value = await get_attribute(conn, element, "selected")
            return get_attribute_value.lower() == ""
        except WebDriverError:
            pass
        return False
    except Exception as e:
        raise WebDriverError("Failed to check if element is selected.") from e


async def get_css_value(conn: AsyncCDPConnection, element, property_name) -> str:
    """Get CSS value"""
    try:
        styles = await get_attribute(conn, element, "style")
        if styles is None:
            return ""
        styles = styles.split(";")
        for s in styles:
            items = s.split(":")
            if len(items) == 2:
                if items[0].strip() == property_name:
                    return items[1].strip()
        raise WebDriverError("Failed to get css value.")
    except Exception as e:
        raise WebDriverError("Failed to get css value.") from e


async def get_property(conn: AsyncCDPConnection, element: dom.NodeId, property_name) -> str:
    try:
        result = await get_attribute(conn, element, property_name)
        if result is None:
            return ""
        return result
    except Exception as e:
        raise WebDriverError("Failed to get property.") from e


async def get_cookies(conn: AsyncCDPConnection) -> list:
    """Get the page cookies"""
    try:
        cookies = await conn.execute(storage.get_cookies())
        return [c.to_json() for c in cookies]
    except Exception as e:
        raise WebDriverError("Failed to get page cookies.") from e


async def add_cookie(conn: AsyncCDPConnection, cookie: Dict[str, Any]):
    """Add cookie

    This function adds a cookie to the WebDriver session.
    Based on W3C WebDriver Specification.

    Args:
        conn: The CDP connection object
        cookie: A dictionary representing the cookie to add
        Example:{
            "domain": ".example.org",
            "httpOnly": True,
            "name": "NID",
            "path": "/",
            "sameSite": "Lax",
            "secure": True,
            "value": "523=Sc0_gs..."
        }
    Returns:
        None

    """
    same_time = (
        network.CookieSameSite.from_json(cookie.get("sameSite")) if cookie.get("sameSite") else None
    )
    try:
        await conn.execute(
            network.set_cookie(
                name=cookie.get("name"),
                value=cookie.get("value"),
                domain=cookie.get("domain"),
                path=cookie.get("path", "/"),
                secure=cookie.get("secure", False),
                http_only=cookie.get("httpOnly", False),
                same_site=same_time,
                expires=cookie.get("expiry"),
            )
        )
    except Exception as e:
        raise WebDriverError("Failed to add cookie.") from e


async def delete_cookie(
    conn: AsyncCDPConnection, name: str, url: Optional[str] = None, domanin: Optional[str] = None
):
    """
    Delete cookie by name

    This function deletes a cookie with the specified name from the WebDriver session.
    Based on W3C WebDriver Specification.

    Args:
        conn: CDPConnection: The connection to the websocket
        name: The name of the cookie to delete

    Returns:
        None

    Raises:
        WebDriverError: If the cookie deletion fails
    """
    try:
        await conn.execute(network.delete_cookies(name=name, url=url, domain=domanin))
    except Exception as e:
        raise WebDriverError(f"Failed to delete cookie '{name}'.") from e


async def refresh_page(conn: AsyncCDPConnection):
    """
    Refreshes the current page by making an HTTP POST request to the server URL.

    Args:
        conn: (CDPConnection): The base URL of the server.
    Returns:
        None
    """
    try:
        await conn.execute(page.reload())
        await GlobalValues.get_document_node(refresh=True)
    except Exception as e:
        raise WebDriverError("Failed to refresh page.") from e


async def go_forward(conn: AsyncCDPConnection):
    """
    Go to page forward.

    This function sends a POST request to the specified URL,
    with an empty payload, and returns True if successful.

    Parameters:
        conn: (CDPConnection): The base URL of the server.

    Returns:
        None
    """
    try:
        result = await conn.execute(page.get_navigation_history())
        current = result[0]
        entry = result[1][current + 1]
        await conn.execute(page.navigate_to_history_entry(entry.id_))
        await _refresh_agents(conn)
    except Exception as e:
        raise WebDriverError("Failed to go to page forward.") from e


async def set_window_rectangle(
    conn: AsyncCDPConnection,
    width: int,
    height: int,
    left: Optional[int] = None,
    top: Optional[int] = None,
):
    """
    Set window rectangle.

    This function sets the window size and position based on W3C WebDriver Specification.

    Args:
        conn: CDPConnection: The connection to the websocket
        width: The desired window width in pixels
        height: The desired window height in pixels
        left: The desired window x coordinate
        top: The desired window y coordinate

    Returns:
        None

    Raises:
        WebDriverError: If the window rectangle setting fails
    """
    try:
        window_for_target = await conn.execute(browser.get_window_for_target())
        window_id = window_for_target[0]
        bounds: Dict[str, Any] = {}
        if left is not None:
            bounds["left"] = left
        if top is not None:
            bounds["top"] = top
        bounds = browser.Bounds.from_json({"width": width, "height": height})
        await conn.execute(browser.set_window_bounds(window_id=window_id, bounds=bounds))
    except Exception as e:
        raise WebDriverError("Failed to set window rectangle.") from e


async def fullscreen_window(conn: AsyncCDPConnection):
    """
    Fullscreen window.

    This function fullscreens the window based on W3C WebDriver Specification.

    Args:
        conn: CDPConnection: The connection to the websocket

    Returns:
        None

    Raises:
        WebDriverError: If the fullscreen operation fails
    """
    try:
        await _set_screen(conn, "normal")
        await _set_screen(conn, "fullscreen")
    except Exception as e:
        raise WebDriverError("Failed to fullscreen window.") from e


async def minimize_window(conn: AsyncCDPConnection):
    """
    Minimize window.

    This function minimizes the window based on W3C WebDriver Specification.

    Args:
        conn: CDPConnection: The connection to the websocket

    Returns:
        None

    Raises:
        WebDriverError: If the minimize operation fails
    """
    try:
        await _set_screen(conn, "normal")
        await _set_screen(conn, "minimized")
    except Exception as e:
        raise WebDriverError("Failed to minimize window.") from e


async def maximize_window(conn: AsyncCDPConnection):
    """
    Maximize window.

    This function maximizes the window based on W3C WebDriver Specification.

    Args:
        conn: CDPConnection: The connection to the websocket

    Returns:
        None

    Raises:
        WebDriverError: If the maximize operation fails
    """
    try:
        await _set_screen(conn, "normal")
        await _set_screen(conn, "maximized")
    except Exception as e:
        raise WebDriverError("Failed to maximize window.") from e


async def switch_to_window(conn: AsyncCDPConnection, handle: target.TargetInfo):
    """
    Switch to window.

    This function switches the WebDriver context to a different window by its handle.
    Based on W3C WebDriver Specification.

    Args:
        conn: CDPConnection: The connection to the websocket.

    Returns:
        new_conn (CDPConnection): The new connection to the websocket.

    Raises:
        WebDriverError: If the switch to window operation fails.
    """
    if not handle:
        raise WebDriverError("Handle not informed")
    try:
        handles = await get_window_handles(conn)
        index = 0
        for h in handles:
            if h.target_id == handle.target_id:
                break
            index += 1
        port = urlparse(conn.url).port
        with urllib.request.urlopen(f"http://127.0.0.1:{port}/json") as r:
            targets = json.loads(r.read())
            targets_page = [p for p in targets if p.get("type") == "page"]
            targets_page.reverse()
            ws_url = targets_page[index]["webSocketDebuggerUrl"]
            await conn.set_url(ws_url)
            await conn.connect()
            new_conn = conn
            await _refresh_agents(new_conn)
            return new_conn
    except Exception as e:
        raise WebDriverError("Failed to switch to window.") from e


async def new_window(conn: AsyncCDPConnection) -> None:
    """
    Open a new window.

    This function opens a new window or tab based on W3C WebDriver Specification.

    Args:
        conn: CDPConnection: The connection to the websocket
        window_type: The type of window to open ('tab' or 'window'). Defaults to 'tab'

    Returns:
        None

    Raises:
        WebDriverError: If the window creation fails
    """
    try:
        await conn.execute(
            target.create_target(
                url="about:blank",
            )
        )
    except Exception as e:
        raise WebDriverError("Failed to open window.") from e


async def switch_to_parent_frame(conn: AsyncCDPConnection):
    """
    Switch to parent frame of 'element_frame'.

    This function switches the WebDriver context to the parent frame of the specified frame element.
    Based on W3C WebDriver Specification.

    Args:
        conn: CDPConnection: The connection to the websocket
        element_frame: The frame element identifier whose parent frame to switch to

    Returns:
        None

    Raises:
        WebDriverError: If the switch to parent frame operation fails
    """
    try:
        doc = await conn.execute(dom.get_document(depth=1))
        doc = await GlobalValues.get_document_node()
        return doc.node_id
    except Exception as e:
        raise WebDriverError("Failed to switch to parent frame.") from e


async def switch_to_frame(conn: AsyncCDPConnection, element_frame: dom.NodeId) -> dom.NodeId:
    """
    Switch to frame 'element_frame'.

    This function switches the WebDriver context to the specified frame element.
    Based on W3C WebDriver Specification.

    Args:
        conn: CDPConnection: The connection to the websocket
        element_frame: The frame element identifier to switch to

    Returns:
        cdp.dom.Node

    Raises:
        WebDriverError: If the switch to frame operation fails
    """
    try:
        desc = await conn.execute(dom.describe_node(node_id=element_frame, depth=1))
        return desc.content_document.node_id
    except Exception as e:
        raise WebDriverError("Failed to switch to frame.") from e


async def delete_all_cookies(conn: AsyncCDPConnection):
    """
    Delete all cookies for the current session.

    This function removes all cookies associated with the active session,
    following the W3C WebDriver Specification for cookie management.

    Args:
        conn: CDPConnection: The connection to the websocket.

    Returns:
        None

    Raises:
        WebDriverError: If the cookie deletion request fails or an error occurs
                        during the deletion process.
    """
    try:
        await conn.execute(network.clear_browser_cookies())
    except Exception as e:
        raise WebDriverError("Failed to delete cookies.") from e


async def get_alert_text(conn: AsyncCDPConnection, element: dom.NodeId) -> str:
    """Get the text from an alert"""
    try:
        return await _handle_alert(conn, element, timeout=TIMEOUT, function_callback=_accept_alert)
    except Exception as e:
        raise WebDriverError("Failed to get the alert text.") from e


async def send_alert_text(
    conn: AsyncCDPConnection, alert_element: dom.NodeId, text, timeout: float = TIMEOUT
):
    """
    Send text to an alert dialog.

    This function sends text to the currently open alert dialog and closes it.
    Based on W3C WebDriver Specification.

    Args:
        conn: CDPConnection: The connection to the websocket
        alert_element: the alert element to send text
        text: The text to send to the alert dialog

    Returns:
        None

    Raises:
        WebDriverError: If sending text to the alert fails
    """
    try:
        await _handle_alert(conn, alert_element, timeout, _send_test_to_prompt_alert, text)
    except Exception as e:
        raise WebDriverError("Failed to sent text to alert.") from e


async def accept_alert(
    conn: AsyncCDPConnection, alert_element: dom.NodeId, timeout: float = TIMEOUT
):
    """
    Accept alert.

    This function accepts the currently open alert dialog.
    Based on W3C WebDriver Specification.

    Args:
        conn: CDPConnection: The connection to the websocket
        alert_element: the alert to be accepted

    Returns:
        None

    Raises:
        WebDriverError: If the alert acceptance fails
    """
    try:
        await _handle_alert(conn, alert_element, timeout, _accept_alert)
    except Exception as e:
        raise WebDriverError("Failed to accept alert.") from e


async def dismiss_alert(
    conn: AsyncCDPConnection, alert_element: dom.NodeId, timeout: float = TIMEOUT
):
    """Dismiss alert

    This function dismisses the currently open alert dialog.
    Based on W3C WebDriver Specification.

    Args:
        conn: CDPConnection: The connection to the websocket.
        alert_element: the alert to be dismissed

    Returns:
        None

    Raises:
        WebDriverError: If the alert dismissal fails.
    """
    try:
        await _handle_alert(conn, alert_element, timeout, _dismiss_alert)
    except Exception as e:
        raise WebDriverError("Failed to dismiss alert.") from e


async def take_screenshot_element(
    conn: AsyncCDPConnection,
    element,
    path="/tmp",
    file_name="caqui",
):
    """Take screenshot of element

    Args:
        conn: CDPConnection: The connection to the websocket..
        element: The identifier of the element to take a screenshot of.
        path: The directory path where the screenshot will be saved.
        file_name: The name of the file to save the screenshot as.

    Returns:
        None

    Raises:
        WebDriverError: If taking the screenshot fails.
    """
    try:
        box_model = await conn.execute(dom.get_box_model(node_id=element))
        content_box = box_model.content[
            0:8
        ]  # Get the 8 points (x1, y1, x2, y2... etc) of the content box

        # Determine the clipping region coordinates from the box model points
        # The content_box provides points of a polygon, we need min/max x and y for the clip
        # Simplified assumption for a simple rectangle:
        x_coords = content_box[::2]
        y_coords = content_box[1::2]
        clip_x = min(x_coords)
        clip_y = min(y_coords)
        clip_width = max(x_coords) - clip_x
        clip_height = max(y_coords) - clip_y

        # Define the Clip object
        clip = page.Viewport(
            x=clip_x,
            y=clip_y,
            width=clip_width,
            height=clip_height,
            scale=1.0,  # Use 1.0 for a standard scale
        )

        now = datetime.datetime.now()
        timestamp = now.strftime(TIME_FORMAT_MICROSECONDS)

        # Capture the screenshot with the specified clip
        screenshot_data = await conn.execute(page.capture_screenshot(clip=clip))
        save_picture(timestamp, path, file_name, screenshot_data)
    except Exception as e:
        raise WebDriverError("Failed to take screenshot from element.") from e


async def take_screenshot(
    conn: AsyncCDPConnection,
    path: str = "/tmp",
    file_name: str = "caqui",
):
    """Take screenshot

    Args:
        conn: CDPConnection: The connection to the websocket..
        path: The directory path where the screenshot will be saved.
        file_name: The name of the file to save the screenshot as.

    Returns:
        NOne

    Raises:
        WebDriverError: If taking the screenshot fails.
    """
    try:
        # 2. Get the full layout metrics of the page
        # This command returns the content size of the page, regardless of the current viewport
        metrics = await conn.execute(page.get_layout_metrics())
        content_size = metrics[-1]
        full_width = content_size.width
        full_height = content_size.height

        # 3. Override the device metrics to match the full page size
        # This effectively makes the entire page visible in a single "viewport"
        viewport = page.Viewport(0, 0, full_width, full_height, 1)
        await conn.execute(
            emulation.set_device_metrics_override(
                width=full_width,
                height=full_height,
                device_scale_factor=1,
                mobile=False,
                screen_width=full_width,
                screen_height=full_height,
                position_x=0,
                position_y=0,
                viewport=viewport,
            )
        )

        screenshot_data = await conn.execute(page.capture_screenshot(quality=90))

        # 5. Reset viewport emulation (optional, good practice)
        await conn.execute(emulation.clear_device_metrics_override())

        # 6. Save the screenshot data to a file
        now = datetime.datetime.now()
        timestamp = now.strftime(TIME_FORMAT_MICROSECONDS)
        save_picture(timestamp, path, file_name, screenshot_data)
    except Exception as e:
        raise WebDriverError("Failed to take screenshot.") from e


async def get_named_cookie(conn, name) -> dict:
    """Get cookie by name.

    This function retrieves a cookie from the WebDriver session based on the specified name.

    Args:
        conn: CDPConnection: The connection to the websocket..
        name: The name of the cookie to retrieve.

    Returns:
        A dictionary representing the cookie if found, otherwise an empty dictionary.

    Raises:
        WebDriverError: If the request to get the cookie fails.
    """
    try:
        cookies = await get_cookies(conn)
        return [c for c in cookies if c.get("name") == name][0]
    except Exception as e:
        raise WebDriverError(f"Failed to get cookie '{name}'.") from e


async def get_computed_label(conn: AsyncCDPConnection, element: dom.NodeId) -> str:
    """Get the element tag computed label. Get the accessibility name.

    Args:
        conn: CDPConnection: The connection to the websocket..
        element: The identifier of the element to retrieve the computed label for.

    Returns:
        The computed label of the element, which represents its accessibility name.

    Raises:
        WebDriverError: If retrieving the computed label fails.
    """
    try:
        # Query the accessibility tree for the specific node
        # The queryAXTree command takes a nodeId (among other possible identifiers)
        ax_nodes_response = await conn.execute(accessibility.get_partial_ax_tree(node_id=element))

        # The response contains a list of AXNode objects
        # The target node's info should be in the list
        for ax_node in ax_nodes_response:
            # Check if this is the correct node and extract its name/label
            # The 'name' property holds the computed accessible name
            if ax_node.name.value:
                return ax_node.name.value
        return ""
    except Exception as e:
        raise WebDriverError("Failed to get element computed label.") from e


async def get_computed_role(conn: AsyncCDPConnection, element: dom.NodeId) -> str:
    """Get the element tag computed role (the element role).

    Args:
        conn: CDPConnection: The connection to the websocket..
        element: The identifier of the element to retrieve the computed role for.

    Returns:
        A string representing the computed role

    Raises:
        WebDriverError: If retrieving the computed role fails.
    """
    try:
        # Get accessibility information for the node, including the role
        # We use getAXNodeAndAncestors to retrieve the AXNode object which holds computed properties
        ax_nodes_response = await conn.execute(
            accessibility.get_ax_node_and_ancestors(node_id=element)
        )

        # The response is a list of nodes, the first one is the target node
        if ax_nodes_response:
            target_ax_node = ax_nodes_response[0]
            computed_role = target_ax_node.role.value
            return computed_role
        return ""
    except Exception as e:
        raise WebDriverError("Failed to get element computed label.") from e


async def get_tag_name(conn: AsyncCDPConnection, element: dom.NodeId) -> str:
    """
    Get the tag name of a specified element in a WebDriver session.

    Parameters:
        conn: CDPConnection: The connection to the websocket.
        element: The identifier for the specific element whose tag name is to be retrieved.

    Returns:
        A string representing the tag name of the specified element.

    Raises:
        WebDriverError: If there is an error while attempting to retrieve the element's tag name.
    """
    try:
        result = await conn.execute(dom.describe_node(node_id=element))
        return result.node_name
    except Exception as e:
        raise WebDriverError("Failed to get element name.") from e


# async def get_shadow_element_v1(
#     conn: CDPConnection,
#     shadow_root_locator_type: str,
#     shadoe_root_locator_value: str,
#     locator_type: str,
#     locator_value: str,
# ) -> dom.NodeId:
#     """
#     Get the shadow root element from a web page using the W3C WebDriver Specification.

#     Parameters:
#         conn: CDPConnection: The connection to the websocket.
#         shadow_element: The ID or name of the shadow element to retrieve.
#         locator_type: The type of locator to use (e.g., 'css selector', 'xpath').
#         locator_value: The value of the locator to find the element.

#     Returns:
#         The shadow root element as a string, or an empty string if not found.

#     Raises:
#         WebDriverError: If there is an error retrieving the shadow element.
#     """
#     try:
#         host_node = await _find_element(
#             conn, shadow_root_locator_type, shadoe_root_locator_value, depth=-1, pierce=True
#         )
#         host_details = await conn.execute(dom.describe_node(node_id=host_node, depth=1))
#         shadow_roots = host_details.shadow_roots
#         return await _find_element(conn, locator_type, locator_value, shadow_roots[0].node_id)
#     except Exception as e:
#         raise WebDriverError("Failed to get the element shadow.") from e


async def get_shadow_element(
    conn: AsyncCDPConnection,
    shadow_element,
    locator_type: str,
    locator_value: str,
) -> dom.NodeId:
    """
    Get the shadow root element from a web page using the W3C WebDriver Specification.

    Parameters:
        conn: CDPConnection: The connection to the websocket.
        shadow_element: The ID or name of the shadow element to retrieve.
        locator_type: The type of locator to use (e.g., 'css selector', 'xpath').
        locator_value: The value of the locator to find the element.

    Returns:
        The shadow root element as a string, or an empty string if not found.

    Raises:
        WebDriverError: If there is an error retrieving the shadow element.
    """
    try:
        host_details = await conn.execute(dom.describe_node(node_id=shadow_element, depth=1))
        shadow_roots = host_details.shadow_roots
        return await _find_element(conn, locator_type, locator_value, shadow_roots[0].node_id)
    except Exception as e:
        raise WebDriverError("Failed to get the element shadow.") from e


async def get_shadow_elements(
    conn: AsyncCDPConnection,
    shadow_element: dom.NodeId,
    locator_type: str,
    locator_value: str,
) -> List[dom.NodeId]:
    """
    Get the list of shadow root elements.

    Args:
        conn: CDPConnection: The connection to the websocket.
        shadow_element: The identifier for the shadow element to retrieve.
        locator_type: The type of locator to use (e.g., 'css selector', 'xpath').
        locator_value: The value of the locator to find the element.

    Returns:
        A list of shadow root element identifiers.

    Raises:
        WebDriverError: If there is an error retrieving the shadow elements.
    """
    """Get the list of shadow root elements"""
    try:
        host_details = await conn.execute(dom.describe_node(node_id=shadow_element, depth=1))
        shadow_roots = host_details.shadow_roots
        return await _find_all_elements(conn, locator_type, locator_value, shadow_roots[0].node_id)
    except Exception as e:
        raise WebDriverError("Failed to get the element shadow.") from e


async def get_rect(conn: AsyncCDPConnection, element) -> dict:
    """Get the element rectangle"""
    try:
        result = await conn.execute(dom.get_box_model(node_id=element))
        # Quad([205, 103.4375, 226.484375, 103.4375, 226.484375, 118.4375, 205, 118.4375])
        content = result.content
        height = abs(content[-1] - content[1])
        width = abs(content[2] - content[0])
        x = content[0]
        y = content[1]
        return {"height": height, "width": width, "x": x, "y": y}
    except Exception as e:
        raise WebDriverError("Failed to get element rect.") from e


async def actions_move_to_element(conn: AsyncCDPConnection, element: dom.NodeId) -> None:
    """
    Move to an element simulating a mouse movement.

    This function sends a WebDriver Actions command to move the mouse pointer to a
    specified element, following the W3C WebDriver Specification.

    Args:
        conn: CDPConnection: The connection to the websocket..
        element: The element identifier (W3C element reference) to move the pointer to.

    Returns:
        None

    Raises:
        WebDriverError: If the action fails to move to the element.

    Note:
        The mouse movement is performed with zero duration, resulting in an instant
        pointer movement to the element's location. The element reference follows the
        W3C WebDriver specification format.
    """
    """Move to an element simulating a mouse movement"""
    try:
        x, y = await _get_element_center(conn, element)
        await conn.execute(
            input_.dispatch_mouse_event(
                "mouseMoved",
                x=x,
                y=y,
            )
        )
    except Exception as e:
        raise WebDriverError("Failed to move to element.") from e


async def actions_scroll_to_element(
    conn: AsyncCDPConnection,
    element,
):
    """Scroll to an element simulating a mouse movement"""
    try:
        remote_object = await conn.execute(dom.resolve_node(node_id=element))
        await conn.execute(
            runtime.call_function_on(
                function_declaration="""
                    function() {
                        this.scrollIntoView({
                            block: 'center',
                            inline: 'center',
                            behavior: 'instant'
                        });
                    }""",
                object_id=remote_object.object_id,
            )
        )
    except Exception as e:
        raise WebDriverError("Failed to scroll to element.") from e


async def submit(conn: AsyncCDPConnection, element):
    """Submit a form. It is similar to 'submit' funtion in Seleniu
    It is not part of W3C WebDriver. Just added for convenience
    """
    try:
        remote_object = await conn.execute(dom.resolve_node(node_id=element))
        await conn.execute(
            runtime.call_function_on(
                function_declaration="""
                    function() {
                        this.submit();
                    }""",
                object_id=remote_object.object_id,
            )
        )
    except Exception as e:
        raise WebDriverError("Failed to submit form.") from e


async def find_children_elements(
    conn: AsyncCDPConnection,
    parent_element: dom.NodeId,
    locator_type: str,
    locator_value: str,
):
    """Find the children elements by 'locator_type'

    If the 'parent_element' is a shadow element, set the 'locator_type' as 'id' or
    'css selector'
    """
    try:
        return await _find_all_elements(conn, locator_type, locator_value, parent_element)
    except Exception as e:
        raise WebDriverError(
            f"Failed to find the children elements from '{parent_element}'."
        ) from e


async def find_child_element(
    conn: AsyncCDPConnection,
    parent_element: dom.NodeId,
    locator_type: str,
    locator_value: str,
):
    """Find the child element by 'locator_type'"""
    try:
        return await find_element(conn, locator_type, locator_value, parent_element)
    except Exception as e:
        raise WebDriverError(f"Failed to find the child element from '{parent_element}'.") from e


async def get_page_source(conn: AsyncCDPConnection) -> str:
    """Get the page source (all content)"""
    try:
        doc = await GlobalValues.get_document_node()
        return await conn.execute(dom.get_outer_html(node_id=doc.node_id))
    except Exception as e:
        raise WebDriverError("Failed to get the page source.") from e


async def execute_script(
    conn: AsyncCDPConnection,
    element: dom.NodeId,
    script: str,
    positive=True,
    return_by_value=False,
):
    """Executes a script, like 'style.background='#000000''
    to change the background of the element"""
    positive = "" if positive else "!"
    try:
        remote_object = await conn.execute(dom.resolve_node(node_id=element))
        return await conn.execute(
            runtime.call_function_on(
                function_declaration="""
                    function() {"""
                f"""
                        return {positive}this.{script};
                    """
                """
                    }
                """,
                object_id=remote_object.object_id,
                return_by_value=return_by_value,
            )
        )
    except Exception as e:
        raise WebDriverError("Failed to execute script.") from e


async def get_active_element(conn: AsyncCDPConnection):
    """Get the active element"""
    try:
        eval_result = await conn.execute(runtime.evaluate(expression="document.activeElement"))
        object_id = eval_result[0].object_id
        if not object_id:
            raise WebDriverError("No active element or it is the document body.")
        node_id_response = await conn.execute(dom.request_node(object_id=object_id))
        return node_id_response
    except Exception as e:
        raise WebDriverError("Failed to check if element is selected.") from e


async def clear_element(conn: AsyncCDPConnection, element: dom.NodeId):
    """Clear the element text"""
    try:
        resolved = await conn.execute(dom.resolve_node(node_id=element))

        await conn.execute(
            runtime.call_function_on(
                object_id=resolved.object_id,
                function_declaration="""
                    function() {
                        this.value = '';
                        this.dispatchEvent(new Event('input', { bubbles: true }));
                        this.dispatchEvent(new Event('change', { bubbles: true }));
                    }
                """,
            )
        )
    except Exception as e:
        raise WebDriverError("Failed to clear the element text.") from e


async def is_element_enabled(conn: AsyncCDPConnection, element) -> bool:
    """Check if element is enabled"""
    try:
        script = "disabled"
        result = await execute_script(conn, element, script, positive=False, return_by_value=True)
        return bool(result[0].value)
    except Exception as e:
        raise WebDriverError("Failed to check if element is enabled.") from e


async def get_window_rectangle(conn: AsyncCDPConnection) -> dict:
    """Get window rectangle"""
    try:
        _, bounds = await conn.execute(browser.get_window_for_target())
        return {
            "x": bounds.left,
            "y": bounds.top,
            "width": bounds.width,
            "height": bounds.height,
        }
    except Exception as e:
        raise WebDriverError("Failed to get window rectangle.") from e


async def get_window_handles(conn: AsyncCDPConnection) -> List[target.TargetInfo]:
    """Get window handles"""
    try:
        targets = await conn.execute(target.get_targets())
        return [t for t in targets if t.type_ == "page"]
    except Exception as e:
        raise WebDriverError("Failed to get window handles.") from e


async def close_window(conn: AsyncCDPConnection):
    """Close active window"""
    try:
        await new_window(conn)
        await conn.execute(target.page.close())
    except Exception as e:
        raise WebDriverError("Failed to close active window.") from e


async def get_window(conn) -> str:
    """Get window"""
    try:
        result = await get_window_handles(conn)
        return result[0].target_id
    except Exception as e:
        raise WebDriverError("Failed to get window.") from e


async def get_status(conn: AsyncCDPConnection) -> dict:
    """Returns the status and details of the WebDriver"""
    try:
        await conn.execute(target.get_targets())
        return {"ready": True}
    except Exception as e:
        raise WebDriverError("Failed to get status.") from e



# Copyright (C) 2023 Caqui - All Rights Reserved
# You may use, distribute and modify this code under the
# terms of the MIT license.
# Visit: https://github.com/douglasdcm/caqui

import os
from typing import List

from caqui.cdp.by import By
from caqui.cdp.engine import synchronous
from caqui.cdp.synchronous.action_chains import ActionChains
from caqui.cdp.synchronous.alert import Alert
from caqui.cdp.synchronous.element import Element
from caqui.cdp.synchronous.switch_to import SwitchTo
from caqui.cdp.synchronous.window import Window
from caqui.helper import deprecated

TIMEOUT = 120  # seconds


class SyncDriverCDP:
    _instance = None

    def __init__(
        self,
        conn,
    ) -> None:
        """Mimics Selenium methods

        Args:
            conn: a client to make HTTP requests
        """
        self._conn = conn
        self._current_element = None

    @property
    def conn(self) -> str:
        """Returns tne websocket connect"""
        return self._conn

    @property
    def window(self) -> Window:
        """Returns the current `Window` object"""
        return Window(self)

    @property
    def actions(self) -> ActionChains:
        """Returns the `ActionChains` object"""
        return ActionChains(self)

    @property
    def alert(self) -> Alert:
        """Returns the `Alert` object"""
        return Alert(self)

    @property
    def switch_to(self) -> SwitchTo:
        """Returns the `SwithTo` object"""
        return SwitchTo(self)

    # TODO test it
    def get_current_window_handle(self) -> str:
        """Returns the current window handle"""
        return synchronous.get_window(self._conn)

    def get_page_source(self) -> str:
        return synchronous.get_page_source(self._conn)

    def get_window_handles(self) -> str:
        """Returns the text of the element"""
        return synchronous.get_window_handles(
            self._conn,
        )

    def get_title(self) -> str:
        """Returns the text of the element"""
        return synchronous.get_title(
            self._conn,
        )

    def get_current_url(self) -> str:
        """Returns the text of the element"""
        return synchronous.get_url(
            self._conn,
        )

    def close(self) -> None:
        """Closes the window"""
        synchronous.close_window(
            self._conn,
        )

    def execute_script(self, script: str, args: List = []):
        """
        Execute a JavaScript script on the browser.

        Args:
            script (str): The JavaScript script to execute.
            args (list[str], optional): Variable arguments for the script. Defaults to [].

        Returns:
            result: The result of the executed script.
        """
        element = synchronous.find_element(self._conn, By.XPATH, "//body")
        synchronous.execute_script(
            self._conn,
            element,
            script,
        )

    # TODO test it
    def set_window_position(self, x: int, y: int) -> None:
        """Repositions the page"""
        rect = synchronous.get_window_rectangle(
            self._conn,
        )
        return synchronous.set_window_rectangle(
            self._conn,
            rect.get("width", 0),
            rect.get("height", 0),
            x,
            y,
        )

    def set_window_size(self, width: int, height: int) -> None:
        """Resizes the page"""
        return synchronous.set_window_rectangle(
            self._conn,
            width,
            height,
            0,
            0,
        )

    # TODO test it
    def get_window_position(self):
        """Returns the window rectangle"""
        return synchronous.get_window_rectangle(
            self._conn,
        )

    def get_window_size(self) -> dict:
        """Returns the window rectangle"""
        return synchronous.get_window_rectangle(
            self._conn,
        )

    def save_screenshot(self, file: str) -> bool:
        """Takes a scheenshot of the page"""
        path = os.path.dirname(file)
        if not path:
            path = "./"
        file_name = os.path.basename(file)
        return synchronous.take_screenshot(
            self._conn,
            path,
            file_name,
        )

    def delete_all_cookies(self) -> None:
        """Deletes all storaged cookies"""
        synchronous.delete_all_cookies(
            self._conn,
        )

    def delete_cookie(self, cookie_name) -> None:
        """Delete the desired cookie"""
        cookie = synchronous.get_named_cookie(self._conn, cookie_name)
        synchronous.delete_cookie(self._conn, cookie_name, cookie.get("url"), cookie.get("domain"))

    def get_cookies(self) -> List[dict]:
        """Get all cookies"""
        return synchronous.get_cookies(
            self._conn,
        )

    def get_cookie(self, cookie_name: str) -> dict:
        """Get the desired cookie"""
        return synchronous.get_named_cookie(
            self._conn,
            cookie_name,
        )

    def add_cookie(self, cookie: dict) -> None:
        """Add a new cookie"""
        synchronous.add_cookie(
            self._conn,
            cookie,
        )

    @deprecated
    def implicitly_wait(self, timeouts: int) -> None:
        """Set implicty timeouts.`Note` Present for backward-compatibility only"""
        pass

    def back(self) -> None:
        """This command causes the browser to traverse one step backward
        in the joint session history of the
        current browse. This is equivalent to pressing the back button in the browser."""
        synchronous.go_back(
            self._conn,
        )

    def forward(self) -> None:
        """Go page forward"""
        synchronous.go_forward(
            self._conn,
        )

    def refresh(self) -> None:
        """Refreshs the page"""
        synchronous.refresh_page(
            self._conn,
        )

    def fullscreen_window(self) -> None:
        """Sets the page in fullscreen"""
        synchronous.fullscreen_window(
            self._conn,
        )

    def minimize_window(self) -> None:
        """Minimizes the page"""
        synchronous.minimize_window(
            self._conn,
        )

    def maximize_window(self) -> None:
        """Maximizes the page"""
        synchronous.maximize_window(
            self._conn,
        )

    def get(self, url: str) -> None:
        """Navigates to URL `url`"""
        synchronous.get(
            self._conn,
            url,
        )

    def find_elements(self, locator: str, value: str) -> List[Element]:
        """Search the DOM elements by 'locator', for example, 'xpath'"""
        elements = synchronous.find_elements(self._conn, locator, value)
        return [Element(e, self) for e in elements]

    def find_element(self, locator: str, value: str) -> Element:
        """Find an element by a 'locator', for example 'xpath'

        `Attention`: it opens Alerts/Prompt elements automatically
        Not applicable for elements in iframes.
        """
        element = synchronous.find_element(self._conn, locator, value)
        self._current_element = Element(element, self)
        return self._current_element



# Copyright (C) 2023 Caqui - All Rights Reserved
# You may use, distribute and modify this code under the
# terms of the MIT license.
# Visit: https://github.com/douglasdcm/caqui

import os
from typing import Callable, Dict, List, Union

from aiohttp import ClientSession

from caqui.constants import Specification
from caqui.exceptions import WebDriverError
from caqui.webdriver.action_chains import ActionChainsJsonWire, ActionChainsW3C
from caqui.webdriver.alert import Alert
from caqui.webdriver.capabilities import BaseCapabilitiesBuilder
from caqui.webdriver.element import Element
from caqui.webdriver.engine import asynchronous, synchronous
from caqui.webdriver.switch_to import SwitchToJsonWire, SwitchToW3C
from caqui.webdriver.window import Window

TIMEOUT = 120  # seconds


SWITCHTO_IMPLEMENTATIONS = {
    Specification.FIREFOX: SwitchToW3C,
    Specification.CHROME: SwitchToJsonWire,
    Specification.EDGE: SwitchToJsonWire,
    Specification.OPERA: SwitchToJsonWire,
    Specification.JSONWIRE: SwitchToJsonWire,
    Specification.W3C: SwitchToW3C,
}

ACTION_CHAINS_IMPLEMENTATIONS = {
    Specification.FIREFOX: ActionChainsW3C,
    Specification.CHROME: ActionChainsJsonWire,
    Specification.EDGE: ActionChainsJsonWire,
    Specification.OPERA: ActionChainsJsonWire,
    Specification.JSONWIRE: ActionChainsJsonWire,
    Specification.W3C: ActionChainsW3C,
}


class _FindElement:
    async def find_element(self, async_driver: "AsyncDriver", locator, value) -> Element:
        raise NotImplementedError("Not implemented by subclass")


class _FindElementW3C(_FindElement):
    async def find_elements(self, async_driver: "AsyncDriver", locator, value) -> list:
        """Search the DOM elements by 'locator', for example, 'xpath'"""
        elements = await asynchronous.find_elements(
            async_driver._server_url,
            async_driver._session,
            locator,
            value,
            session_http=async_driver.session_http,
        )
        result = []
        for element in elements:
            el = Element(element, async_driver)
            el.locator = (locator, value)
            result.append(el)
        async_driver._elements_pool.extend(result)
        async_driver._elements_pool = list(set(async_driver._elements_pool))
        return result

    async def find_element(self, async_driver: "AsyncDriver", locator, value) -> Element:
        """Find an element by a 'locator', for example 'xpath'"""
        elements_filtered: List[Element] = [
            e for e in async_driver._elements_pool if e.locator == (locator, value)
        ]
        if elements_filtered:
            return elements_filtered[0]
        element = await asynchronous.find_element(
            async_driver._server_url,
            async_driver._session,
            locator,
            value,
            session_http=async_driver.session_http,
        )
        result = Element(element, async_driver)
        result.locator = (locator, value)
        async_driver._elements_pool.append(result)
        return result


class _FindElementJsonWire:
    async def find_elements(self, async_driver: "AsyncDriver", locator, value) -> List[Element]:
        """Search the DOM elements by 'locator', for example, 'xpath'"""
        elements = await asynchronous.find_elements_jsonwire(
            async_driver._server_url,
            async_driver._session,
            locator,
            value,
            session_http=async_driver.session_http,
        )
        result = []
        for element in elements:
            el = Element(element, async_driver)
            el.locator = (locator, value)
            result.append(el)
        async_driver._elements_pool.extend(result)
        async_driver._elements_pool = list(set(async_driver._elements_pool))
        return result

    async def find_element(self, async_driver: "AsyncDriver", locator, value) -> Element:
        """Find an element by a 'locator', for example 'xpath'"""
        elements_filtered: List[Element] = [
            e for e in async_driver._elements_pool if e.locator == (locator, value)
        ]
        if elements_filtered:
            return elements_filtered[0]
        element = await asynchronous.find_element_jsonwire(
            async_driver._server_url,
            async_driver._session,
            locator,
            value,
            session_http=async_driver.session_http,
        )
        result = Element(element, async_driver)
        result.locator = (locator, value)
        async_driver._elements_pool.append(result)
        return result


FIND_ELEMENT_IMPLEMENTATIONS: Dict[str, Callable] = {
    Specification.FIREFOX: _FindElementW3C,
    Specification.CHROME: _FindElementJsonWire,
    Specification.EDGE: _FindElementJsonWire,
    Specification.OPERA: _FindElementJsonWire,
    Specification.JSONWIRE: _FindElementJsonWire,
    Specification.W3C: _FindElementW3C,
}


class AsyncDriver:
    _instance = None

    def __init__(
        self,
        server_url: str,
        capabilities: Union[BaseCapabilitiesBuilder, dict] = dict(),
        session_http: Union[ClientSession, None] = None,
        port: int = 9999,
        specification: str = "",
    ) -> None:
        """Mimics Selenium methods

        Args:
            server_url: the URL of the remote server running the driver
            capabilities: the configuration to the driver
            session_http: a client to make HTTP requests
            port: the port where the remote server is running the driver
            specification: the specification the driver follows.
            Allowed values are "w3c" or "jsonwire"
                For example, ChromeDriver follows JsonWire protocol while
                GeckoDriver works with W3C
        """
        if isinstance(capabilities, BaseCapabilitiesBuilder):
            self.browser = capabilities.__class__.__name__
        elif specification.lower() in ["w3c", "jsonwire"]:
            self.browser = specification.lower()
        else:
            raise WebDriverError("No valid specification or capabilities")
        self._port = port
        self.session_http = session_http
        self._capabilities: dict = {}
        if isinstance(capabilities, BaseCapabilitiesBuilder):
            self._capabilities = capabilities.to_dict()
        else:
            self._capabilities = capabilities
        self._server_url: str = server_url
        self._session: str = synchronous.get_session(self._server_url, self._capabilities)
        self._elements_pool: List[Element] = []

    @property
    def server_url(self) -> str:
        """Returns the Driver Server URL"""
        return self._server_url

    @property
    def session(self) -> str:
        """Returns tne session id"""
        return self._session

    @property
    def title(self) -> str:
        """Returns the title of the page"""
        return synchronous.get_title(self._server_url, self._session)

    @property
    def current_url(self) -> str:
        """Returns the current URL of the page"""
        return synchronous.get_url(self._server_url, self._session)

    @property
    def window(self) -> Window:
        """Returns the current `Window` object"""
        return Window(self)

    @property
    def actions(self) -> ActionChainsW3C:
        """Returns the `ActionChains` object"""
        return ACTION_CHAINS_IMPLEMENTATIONS[self.browser](self)

    @property
    def alert(self) -> Alert:
        """Returns the `Alert` object"""
        return Alert(self)

    @property
    def switch_to(self) -> SwitchToW3C:
        """Returns the `SwithTo` object"""
        return SWITCHTO_IMPLEMENTATIONS[self.browser](self)

    @property
    def window_handles(self) -> List[str]:
        """Returns the window handles"""
        return synchronous.get_window_handles(self._server_url, self._session)

    @property
    def current_window_handle(self) -> str:
        """Returns the current window handle"""
        return synchronous.get_window(self._server_url, self._session)

    @property
    def page_source(self) -> str:
        return synchronous.get_page_source(self._server_url, self._session)

    def cleanup_cache(self) -> None:
        """
        Clears the cache by resetting the _elements_pool attribute.

        This method is intended to be called when the cache needs to be cleaned up.
        It does not perform any actual cleaning, but rather resets the pool of elements.
        """
        self._elements_pool = []

    def quit(self) -> None:
        """Closes the session"""
        self.cleanup_cache()
        synchronous.close_session(self._server_url, self._session)

    async def close(self) -> None:
        """Closes the window"""
        self.cleanup_cache()
        await asynchronous.close_window(
            self._server_url, self._session, session_http=self.session_http
        )

    async def execute_script(self, script: str, args: List = []):
        """
        Execute a JavaScript script on the browser.

        Args:
            script (str): The JavaScript script to execute.
            args (list[str], optional): Variable arguments for the script. Defaults to [].

        Returns:
            result: The result of the executed script.
        """
        return await asynchronous.execute_script(
            self._server_url, self._session, script, args, session_http=self.session_http
        )

    async def set_window_position(self, x: int, y: int) -> None:
        """Repositions the page"""
        rect = await asynchronous.get_window_rectangle(
            self._server_url, self._session, session_http=self.session_http
        )
        return await asynchronous.set_window_rectangle(
            self._server_url,
            self._session,
            rect.get("width", 0),
            rect.get("height", 0),
            x,
            y,
            session_http=self.session_http,
        )

    async def set_window_size(self, width: int, height: int) -> None:
        """Resizes the page"""
        rect = await asynchronous.get_window_rectangle(
            self._server_url, self._session, session_http=self.session_http
        )
        return await asynchronous.set_window_rectangle(
            self._server_url,
            self._session,
            width,
            height,
            rect.get("x", 0),
            rect.get("y", 0),
            session_http=self.session_http,
        )

    async def get_window_position(self):
        """Returns the window rectangle"""
        return await asynchronous.get_window_rectangle(
            self._server_url, self._session, session_http=self.session_http
        )

    async def get_window_size(self) -> dict:
        """Returns the window rectangle"""
        return await asynchronous.get_window_rectangle(
            self._server_url, self._session, session_http=self.session_http
        )

    async def save_screenshot(self, file: str) -> bool:
        """Takes a scheenshot of the page"""
        path = os.path.dirname(file)
        if not path:
            path = "./"
        file_name = os.path.basename(file)
        return await asynchronous.take_screenshot(
            self._server_url, self._session, path, file_name, session_http=self.session_http
        )

    async def delete_all_cookies(self) -> None:
        """Deletes all storaged cookies"""
        await asynchronous.delete_all_cookies(
            self._server_url, self._session, session_http=self.session_http
        )

    async def delete_cookie(self, cookie_name) -> None:
        """Delete the desired cookie"""
        await asynchronous.delete_cookie(
            self._server_url, self._session, cookie_name, session_http=self.session_http
        )

    async def get_cookies(self) -> List[dict]:
        """Get all cookies"""
        return await asynchronous.get_cookies(
            self._server_url, self._session, session_http=self.session_http
        )

    async def get_cookie(self, cookie_name: str) -> dict:
        """Get the desired cookie"""
        return await asynchronous.get_named_cookie(
            self._server_url, self._session, cookie_name, session_http=self.session_http
        )

    async def add_cookie(self, cookie: dict) -> None:
        """Add a new cookie"""
        await asynchronous.add_cookie(
            self._server_url, self._session, cookie, session_http=self.session_http
        )

    async def implicitly_wait(self, timeouts: int) -> None:
        """Set implicty timeouts"""
        await asynchronous.set_timeouts(
            self._server_url, self._session, timeouts, session_http=self.session_http
        )

    async def back(self) -> None:
        """This command causes the browser to traverse one step backward
        in the joint session history of the
        current browse. This is equivalent to pressing the back button in the browser."""
        self._elements_pool = []
        await asynchronous.go_back(self._server_url, self._session, session_http=self.session_http)

    async def forward(self) -> None:
        """Go page forward"""
        self._elements_pool = []
        await asynchronous.go_forward(
            self._server_url, self._session, session_http=self.session_http
        )

    async def refresh(self) -> None:
        """Refreshs the page"""
        self._elements_pool = []
        await asynchronous.refresh_page(
            self._server_url, self._session, session_http=self.session_http
        )

    async def fullscreen_window(self) -> None:
        """Sets the page in fullscreen"""
        await asynchronous.fullscreen_window(
            self._server_url, self._session, session_http=self.session_http
        )

    async def minimize_window(self) -> None:
        """Minimizes the page"""
        await asynchronous.minimize_window(
            self._server_url, self._session, session_http=self.session_http
        )

    async def maximize_window(self) -> None:
        """Maximizes the page"""
        await asynchronous.maximize_window(
            self._server_url, self._session, session_http=self.session_http
        )

    async def get(self, url: str) -> None:
        """Navigates to URL `url`"""
        self._elements_pool = []
        await asynchronous.go_to_page(
            self._server_url, self._session, url, session_http=self.session_http
        )

    async def find_elements(self, locator: str, value: str) -> List[Element]:
        """Search the DOM elements by 'locator', for example, 'xpath'"""
        return await FIND_ELEMENT_IMPLEMENTATIONS[self.browser]().find_elements(
            self, locator, value
        )

    async def find_element(self, locator: str, value: str) -> Element:
        """Find an element by a 'locator', for example 'xpath'"""
        return await FIND_ELEMENT_IMPLEMENTATIONS[self.browser]().find_element(self, locator, value)



# Copyright (C) 2023 Caqui - All Rights Reserved
# You may use, distribute and modify this code under the
# terms of the MIT license.
# Visit: https://github.com/douglasdcm/caqui


class By:
    """List of locator strategies"""

    CSS_SELECTOR: str = "css selector"
    ID: str = "id"
    XPATH: str = "xpath"
    NAME: str = "name"
    TAG_NAME: str = "tag name"
    CLASS_NAME: str = "class name"
    LINK_TEXT: str = "link text"
    PARTIAL_LINK_TEXT: str = "partial link text"
    # TODO add TEXT and expose to CDP



# Copyright (C) 2023 Caqui - All Rights Reserved
# You may use, distribute and modify this code under the
# terms of the MIT license.
# Visit: https://github.com/douglasdcm/caqui

import base64
import warnings
from functools import lru_cache

from caqui.by import By
from caqui.constants import ELEMENT_JSONWIRE, ELEMENT_W3C
from caqui.exceptions import WebDriverError


def save_picture(session: str, path: str, file_name: str, response: str) -> None:
    """
    Save a screenshot to a file.
    :param session: session id
    :param path: path to save the file
    :param file_name: name of the file
    :param response: base64 encoded screenshot
    """
    with open(f"{path}/{file_name}-{session}.png", "wb") as f:
        f.write(base64.b64decode((response)))


def get_elements(response: dict) -> list:
    """
    Extract the first value from each item in the 'value' key of a response dictionary.

    Args:
        response (dict): A dictionary containing a 'value' key with a list of dictionaries.

    Returns:
        list: A list containing the first value from each dictionary in the 'value' list.

    Example:
        >>> response = {"value": [{"key1": "val1", "key2": "val2"}, {"key3": "val3"}]}
        >>> get_elements(response)
        ['val1', 'val3']
    """
    values = response.get("value", {})
    return [list(value.values())[0] for value in values]


def get_element(response: dict) -> str:
    """Extract the element identifier from a WebDriver response payload.

    The function expects `response` to be a mapping containing a "value" mapping.
    For Chrome-based WebDriver responses the element id is usually stored under
    the ELEMENT key. For Firefox/Gecko the element id is commonly provided as
    the first value of the inner mapping.

    Args:
        response (dict): A WebDriver response dict with a "value" mapping that
                         contains either the ELEMENT key (Chrome) or another
                         single key whose value is the element id (Firefox).

    Returns:
        str: The extracted element identifier.

    Raises:
        AttributeError: If `response` does not contain a "value" mapping (i.e. when
                        `response.get("value")` returns None), attribute access
                        on `None` will raise.
    """
    value = response.get("value", {})
    # Google Chrome
    element = value.get(ELEMENT_W3C)
    if element:
        return element

    # Firefox
    return list(value.values())[0]


def get_element_jsonwire(response: dict) -> str:
    """Extract the element identifier from a WebDriver response payload.

    The function expects `response` to be a mapping containing a "value" mapping.
    For Chrome-based WebDriver responses the element id is usually stored under
    the ELEMENT key. For Firefox/Gecko the element id is commonly provided as
    the first value of the inner mapping.

    Args:
        response (dict): A WebDriver response dict with a "value" mapping that
                         contains either the ELEMENT key (Chrome) or another
                         single key whose value is the element id (Firefox).

    Returns:
        str: The extracted element identifier.

    Raises:
        AttributeError: If `response` does not contain a "value" mapping (i.e. when
                        `response.get("value")` returns None), attribute access
                        on `None` will raise.
    """
    value = response.get("value", {})
    # Google Chrome
    element = value.get(ELEMENT_JSONWIRE)
    if element:
        return element

    # Firefox
    return list(value.values())[0]


@lru_cache(maxsize=42)
def convert_locator_to_css_selector_or_xpath(locator_type: str, locator_value: str) -> tuple:
    """
    Convert a locator type and value to a CSS selector OR XPATH if possible.

    This function attempts to convert an locator expression to an equivalent
    CSS selector. If the conversion fails, the original locator type and value are
    returned unchanged. Results are cached for performance optimization.

    Args:
        locator_type (str): The type of locator (e.g., "xpath", "name").
        locator_value (str): The locator expression value to convert.

    Returns:
        tuple: A tuple containing (locator_type, locator_value) where locator_type
            is either "css selector" if conversion succeeded, or the original
            locator_type if conversion failed or was not applicable.

    Note:
        This function uses LRU caching with a maximum size of 32 entries to
        optimize repeated conversions of the same locator values.
    """
    if locator_type not in list(
        {
            key: value
            for key, value in By.__dict__.items()
            if not key.startswith("__") and not callable(value)
        }.values()
    ):
        raise WebDriverError(f"Locator {locator_type} not allowed")
    if locator_type.lower() == By.ID:
        locator_value = f"#{locator_value}"
        locator_type = By.CSS_SELECTOR
        return locator_type, locator_value
    if locator_type.lower() == By.CLASS_NAME:
        locator_value = f".{locator_value}"
        locator_type = By.CSS_SELECTOR
        return locator_type, locator_value
    if locator_type.lower() == By.NAME:
        locator_value = f"[name='{locator_value}']"
        locator_type = By.CSS_SELECTOR
        return locator_type, locator_value
    if locator_type.lower() == By.LINK_TEXT:
        locator_value = f"//a[text()='{locator_value}']"
        locator_type = By.XPATH
        return locator_type, locator_value
    if locator_type.lower() == By.PARTIAL_LINK_TEXT:
        locator_value = f"//a[contains(text(),'{locator_value}')]"
        locator_type = By.XPATH
        return locator_type, locator_value
    if locator_type.lower() == By.TAG_NAME:
        locator_value = f"//{locator_value}"
        locator_type = By.XPATH
        return locator_type, locator_value
    return locator_type, locator_value


def deprecated(func):
    """Use this decorator to mark functions as deprecated.
    Every time the decorated function runs, it will emit
    a "deprecation" warning."""
    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn(
            f"Call to a deprecated function '{func.__name__}'.",
            category=DeprecationWarning,
            stacklevel=2
        )
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        return func(*args, **kwargs)
    return new_func



# Copyright (C) 2023 Caqui - All Rights Reserved
# You may use, distribute and modify this code under the
# terms of the MIT license.
# Visit: https://github.com/douglasdcm/caqui

import datetime
import json
import threading
import urllib.request
from typing import Any, Dict, List, Optional
from urllib.parse import urlparse

from bs4 import BeautifulSoup

from caqui._vendor.chrome_devtools_protocol.cdp import (
    accessibility,
    browser,
    dom,
    emulation,
    input_,
    network,
    page,
    runtime,
    storage,
    target,
)
from caqui.cdp.by import By
from caqui.cdp.connection import SyncCDPConnection
from caqui.constants import TIME_FORMAT_MICROSECONDS, TIMEOUT
from caqui.exceptions import WebDriverError
from caqui.helper import convert_locator_to_css_selector_or_xpath, save_picture


class GlobalValues:
    _doc: Optional[dom.Node] = None
    conn: SyncCDPConnection = None

    def get_document_node(refresh=False) -> dom.Node:
        if refresh:
            GlobalValues._doc = None
        if not GlobalValues.conn:
            raise WebDriverError("No connection")
        if not GlobalValues._doc:
            GlobalValues._doc = GlobalValues.conn.execute(dom.get_document(depth=-1, pierce=True))
        return GlobalValues._doc


def _get_element_center(conn, element):
    box_model = conn.execute(dom.get_box_model(node_id=element))
    content_box = box_model.content
    center_x = (content_box[0] + content_box[2]) / 2
    center_y = (content_box[1] + content_box[5]) / 2
    return center_x, center_y


def _find_element_by_xpath(
    conn: SyncCDPConnection,
    xpath: str,
    root_id: Optional[dom.NodeId] = None,
) -> dom.NodeId:
    if root_id:
        remote_root = conn.execute(dom.resolve_node(node_id=root_id))
    else:
        node = GlobalValues.get_document_node()
        remote_root = conn.execute(dom.resolve_node(node_id=node.node_id))
    object_id = remote_root.object_id
    call_res = conn.execute(
        runtime.call_function_on(
            object_id=object_id,
            function_declaration=f"""
            function() {{
                const doc = this.ownerDocument || this;
                const r = doc.evaluate(
                    `{xpath}`,
                    this,
                    null,
                    XPathResult.FIRST_ORDERED_NODE_TYPE,
                    null
                );
                return r.singleNodeValue;
            }}
            """,
            return_by_value=False,
        )
    )
    remote = call_res[0]
    if not remote.object_id:
        raise WebDriverError("XPath returned no element")
    node = conn.execute(dom.request_node(object_id=remote.object_id))
    if node == 0:
        raise WebDriverError("Resolved object is not a DOM node")
    return node


def _find_all_elements_by_xpath(
    conn: SyncCDPConnection,
    xpath: str,
    root_id: Optional[dom.NodeId] = None,
) -> List[dom.NodeId]:
    if root_id:
        remote_root = conn.execute(dom.resolve_node(node_id=root_id))
    else:
        node = GlobalValues.get_document_node()
        remote_root = conn.execute(dom.resolve_node(node_id=node.node_id))

    object_id = remote_root.object_id

    call_res = conn.execute(
        runtime.call_function_on(
            object_id=object_id,
            function_declaration=f"""
            function () {{
                const doc = this.ownerDocument || this;
                const snapshot = doc.evaluate(
                    `{xpath}`,
                    this,
                    null,
                    XPathResult.ORDERED_NODE_SNAPSHOT_TYPE,
                    null
                );

                const results = [];
                for (let i = 0; i < snapshot.snapshotLength; i++) {{
                    results.push(snapshot.snapshotItem(i));
                }}
                return results;
            }}
            """,
            return_by_value=False,
        )
    )

    remote_array = call_res[0]

    if not remote_array.object_id:
        raise WebDriverError("XPath returned no elements")

    properties = conn.execute(
        runtime.get_properties(object_id=remote_array.object_id, own_properties=True)
    )

    nodes: List[dom.NodeId] = []

    for prop in properties[0]:
        if not prop.name.isdigit():
            continue

        remote_node = prop.value
        if not remote_node or not remote_node.object_id:
            continue

        node_id = conn.execute(dom.request_node(object_id=remote_node.object_id))

        if node_id != 0:
            nodes.append(node_id)

    if not nodes:
        raise WebDriverError("Resolved objects are not DOM nodes")

    return nodes


def describe_node_id(conn: SyncCDPConnection, node_id: dom.NodeId):
    if not node_id:
        raise WebDriverError(f"Invalid node id '{node_id}'")
    return conn.execute(dom.describe_node(node_id))


def _find_element_by_css_selector(
    conn: SyncCDPConnection, root_id: dom.NodeId, depth: int, pierce: bool, selector: str
):
    if root_id:
        node = conn.execute(dom.describe_node(root_id))
    else:
        node = GlobalValues.get_document_node()
    node_id = conn.execute(dom.query_selector(node_id=node.node_id, selector=selector))
    if not node_id:
        raise WebDriverError()
    return node_id


def _find_element(
    conn: SyncCDPConnection,
    locator_type,
    locator_value: str,
    root_id: dom.NodeId = None,
    depth=1,
    pierce=False,
):
    locator_type, selector = convert_locator_to_css_selector_or_xpath(locator_type, locator_value)
    if locator_type == By.XPATH:
        try:
            return _find_element_by_xpath(conn, selector, root_id)
        except Exception:
            raise
    if locator_type == By.CSS_SELECTOR:
        return _find_element_by_css_selector(conn, root_id, depth, pierce, selector)
    raise WebDriverError(f"Could not find element with selector: {locator_value}")


def _find_all_elements(
    conn: SyncCDPConnection,
    locator_type,
    locator_value: str,
    root_id: dom.NodeId = None,
):
    locator_type, locator_value = convert_locator_to_css_selector_or_xpath(
        locator_type, locator_value
    )
    if locator_type == By.XPATH:
        return _find_all_elements_by_xpath(conn, locator_value, root_id)
    if locator_type == By.CSS_SELECTOR:
        if root_id:
            node = conn.execute(dom.describe_node(root_id))
        else:
            node = GlobalValues.get_document_node()
        node_id = conn.execute(dom.query_selector_all(node_id=node.node_id, selector=locator_value))
        if not node_id:
            raise WebDriverError()
        return node_id


def _handle_alert(conn: SyncCDPConnection, alert_element, timeout, function_callback, text=None):
    click_task = threading.Thread(target=click, args=(conn, alert_element))
    click_task.daemon = True
    click_task.start()
    click_task.join()
    current_datetime = datetime.datetime.now()
    time_to_add = datetime.timedelta(seconds=timeout)
    new_datetime = current_datetime + time_to_add
    while 1 == 1:  # Used 1==1 and not True to deceive VSCode
        while datetime.datetime.now() < new_datetime:
            event = conn.get_event_nowait()
            if event is None:
                continue
            if isinstance(event, page.JavascriptDialogOpening):
                function_callback(conn, text)
                return event.message
        raise TimeoutError()
    click_task


def _send_test_to_prompt_alert(conn: SyncCDPConnection, text):
    conn.execute(page.handle_java_script_dialog(accept=True, prompt_text=text))


def _accept_alert(conn: SyncCDPConnection, text):
    conn.execute(page.handle_java_script_dialog(accept=True))


def _dismiss_alert(conn: SyncCDPConnection, text):
    conn.execute(page.handle_java_script_dialog(accept=False))


def _set_screen(conn, screen_type: str = "fullscreen"):
    window_for_target = conn.execute(browser.get_window_for_target())
    window_id = window_for_target[0]
    bounds = browser.Bounds.from_json({"windowState": screen_type})
    conn.execute(browser.set_window_bounds(window_id=window_id, bounds=bounds))


def get(conn: SyncCDPConnection, page_url: str) -> None:
    """Does the same of 'go_to_page'. Added to be compatible with selenium method name'"""
    try:
        conn.execute(page.enable())
        conn.execute(page.navigate(url=page_url))
        conn.execute(
            target.set_auto_attach(auto_attach=True, wait_for_debugger_on_start=False, flatten=True)
        )
        _refresh_agents(conn)
    except Exception as e:
        raise WebDriverError(f"Failed to navigate to page '{page_url}'.") from e


def _refresh_agents(conn: SyncCDPConnection):
    conn.execute(page.enable())
    conn.execute(dom.enable())
    conn.execute(runtime.enable())
    conn.execute(network.enable())
    conn.execute(accessibility.enable()),
    GlobalValues.conn = conn
    GlobalValues.get_document_node(refresh=True)


def go_to_page(conn, page_url: str) -> None:
    """Navigate to 'page_url'"""
    get(conn, page_url)


def find_element(
    conn: SyncCDPConnection, locator_type: str, locator_value: str, root_id: dom.NodeId = None
) -> dom.NodeId:
    """Find an element by a 'selector', for example an 'xpath' like '//div[@id="example"]'

    `Attention`: it opens Alerts/Prompt elements automatically.
    Not applicable for elements in iframes.
    """
    try:
        return _find_element(conn, locator_type, locator_value, root_id)
    except Exception as e:
        raise WebDriverError(f"Could not find element with selector: {locator_value}") from e


def find_elements(conn: SyncCDPConnection, locator_type, locator_value: str):
    """Find an element by a 'selector', for example an 'xpath' like '//div[@id="example"]'"""
    try:
        return _find_all_elements(conn, locator_type, locator_value)
    except Exception as e:
        raise WebDriverError(f"Could not find element with selector: {locator_value}") from e


def click(conn: SyncCDPConnection, element):
    """Click on an element"""
    try:
        center_x, center_y = _get_element_center(conn, element)
        commands = []
        mouse_button = input_.MouseButton("left")
        commands.append(
            input_.dispatch_mouse_event(
                "mousePressed", center_x, center_y, button=mouse_button, click_count=1
            )
        )
        commands.append(
            input_.dispatch_mouse_event(
                "mouseReleased", center_x, center_y, button=mouse_button, click_count=1
            )
        )
        for command in commands:
            conn.execute(command, wait=False)
    except Exception as e:
        raise WebDriverError("Failed to click on element.") from e


def send_keys(conn: SyncCDPConnection, element, text: str):
    """Send keys to an element"""
    try:
        conn.execute(dom.focus(node_id=element))
        for char in text:
            conn.execute(
                input_.dispatch_key_event(
                    "keyDown",
                    text=char,
                    unmodified_text=char,
                    windows_virtual_key_code=ord(char),
                    native_virtual_key_code=ord(char),
                )
            )
            conn.execute(
                input_.dispatch_key_event(
                    "keyUp",
                    text=char,
                    unmodified_text=char,
                    windows_virtual_key_code=ord(char),
                    native_virtual_key_code=ord(char),
                )
            )
    except Exception as e:
        raise WebDriverError("Failed to send keys to element.") from e


def get_text(conn: SyncCDPConnection, element) -> str:
    """Get the text of an element"""
    try:
        result = execute_script(conn, element, "value")
        if result and result[0].value:
            return result[0].value
        html_doc = conn.execute(dom.get_outer_html(node_id=element))
        soup = BeautifulSoup(html_doc, "html.parser")
        return soup.get_text()
    except Exception as e:
        raise WebDriverError("Failed to get text from element.") from e


def get_attribute(conn: SyncCDPConnection, element, attribute: str) -> str:
    """Get the given HTML attribute of an element, for example, 'aria-valuenow'"""
    try:
        attributes = conn.execute(dom.get_attributes(node_id=element))
        attr_dict = {attributes[i]: attributes[i + 1] for i in range(0, len(attributes), 2)}
        result = attr_dict.get(attribute)
        if result is None:
            raise WebDriverError("Failed to get value from element.")
        return result
    except Exception as e:
        raise WebDriverError("Failed to get value from element.") from e


def get_title(conn: SyncCDPConnection) -> str:
    """Get the page title"""
    try:
        target_info = conn.execute(target.get_target_info())
        return target_info.title if target_info.type_ == "page" else ""
    except Exception:
        pass
    try:
        root_node = GlobalValues.get_document_node()
        title_node_id = conn.execute(
            dom.query_selector(node_id=root_node.node_id, selector="title")
        )
        result = conn.execute(dom.get_outer_html(node_id=title_node_id))
        return BeautifulSoup(result, "html.parser").get_text()
    except Exception as e:
        raise WebDriverError("Failed to get page title.") from e


def get_url(conn: SyncCDPConnection) -> str:
    """Returns the URL from web page:"""
    try:
        frame_tree = conn.execute(page.get_frame_tree())
        current_url = frame_tree.frame.url
        return current_url
    except Exception as e:
        raise WebDriverError("Failed to get page url.") from e


def go_back(conn: SyncCDPConnection):
    try:
        result = conn.execute(page.get_navigation_history())
        current = result[0]
        if current <= 0:
            return
        entry = result[1][current - 1]
        conn.execute(page.navigate_to_history_entry(entry.id_))
        _refresh_agents(conn)
    except Exception as e:
        raise WebDriverError("Failed to go back to page.") from e


def is_element_selected(conn: SyncCDPConnection, element) -> bool:
    """Check if element is selected"""
    try:
        try:
            get_attribute_value = get_attribute(conn, element, "checked")
            return get_attribute_value.lower() == ""
        except WebDriverError:
            pass
        try:
            get_attribute_value = get_attribute(conn, element, "selected")
            return get_attribute_value.lower() == ""
        except WebDriverError:
            pass
        return False
    except Exception as e:
        raise WebDriverError("Failed to check if element is selected.") from e


def get_css_value(conn: SyncCDPConnection, element, property_name) -> str:
    """Get CSS value"""
    try:
        styles = get_attribute(conn, element, "style")
        if styles is None:
            return ""
        styles = styles.split(";")
        for s in styles:
            items = s.split(":")
            if len(items) == 2:
                if items[0].strip() == property_name:
                    return items[1].strip()
        raise WebDriverError("Failed to get css value.")
    except Exception as e:
        raise WebDriverError("Failed to get css value.") from e


def get_property(conn: SyncCDPConnection, element: dom.NodeId, property_name) -> str:
    try:
        result = get_attribute(conn, element, property_name)
        if result is None:
            return ""
        return result
    except Exception as e:
        raise WebDriverError("Failed to get property.") from e


def get_cookies(conn: SyncCDPConnection) -> list:
    """Get the page cookies"""
    try:
        cookies = conn.execute(storage.get_cookies())
        return [c.to_json() for c in cookies]
    except Exception as e:
        raise WebDriverError("Failed to get page cookies.") from e


def add_cookie(conn: SyncCDPConnection, cookie: Dict[str, Any]):
    """Add cookie

    This function adds a cookie to the WebDriver session.
    Based on W3C WebDriver Specification.

    Args:
        conn: The CDP connection object
        cookie: A dictionary representing the cookie to add
        Example:{
            "domain": ".example.org",
            "httpOnly": True,
            "name": "NID",
            "path": "/",
            "sameSite": "Lax",
            "secure": True,
            "value": "523=Sc0_gs..."
        }
    Returns:
        None

    """
    same_time = (
        network.CookieSameSite.from_json(cookie.get("sameSite")) if cookie.get("sameSite") else None
    )
    try:
        conn.execute(
            network.set_cookie(
                name=cookie.get("name"),
                value=cookie.get("value"),
                domain=cookie.get("domain"),
                path=cookie.get("path", "/"),
                secure=cookie.get("secure", False),
                http_only=cookie.get("httpOnly", False),
                same_site=same_time,
                expires=cookie.get("expiry"),
            )
        )
    except Exception as e:
        raise WebDriverError("Failed to add cookie.") from e


def delete_cookie(
    conn: SyncCDPConnection, name: str, url: Optional[str] = None, domanin: Optional[str] = None
):
    """
    Delete cookie by name

    This function deletes a cookie with the specified name from the WebDriver session.
    Based on W3C WebDriver Specification.

    Args:
        conn: CDPConnectionSync: The connection to the websocket
        name: The name of the cookie to delete

    Returns:
        None

    Raises:
        WebDriverError: If the cookie deletion fails
    """
    try:
        conn.execute(network.delete_cookies(name=name, url=url, domain=domanin))
    except Exception as e:
        raise WebDriverError(f"Failed to delete cookie '{name}'.") from e


def refresh_page(conn: SyncCDPConnection):
    """
    Refreshes the current page by making an HTTP POST request to the server URL.

    Args:
        conn: (CDPConnectionSync): The base URL of the server.
    Returns:
        None
    """
    try:
        conn.execute(page.reload())
        GlobalValues.get_document_node(refresh=True)
    except Exception as e:
        raise WebDriverError("Failed to refresh page.") from e


def go_forward(conn: SyncCDPConnection):
    """
    Go to page forward.

    This function sends a POST request to the specified URL,
    with an empty payload, and returns True if successful.

    Parameters:
        conn: (CDPConnectionSync): The base URL of the server.

    Returns:
        None
    """
    try:
        result = conn.execute(page.get_navigation_history())
        current = result[0]
        entry = result[1][current + 1]
        conn.execute(page.navigate_to_history_entry(entry.id_))
        _refresh_agents(conn)
    except Exception as e:
        raise WebDriverError("Failed to go to page forward.") from e


def set_window_rectangle(
    conn: SyncCDPConnection,
    width: int,
    height: int,
    left: Optional[int] = None,
    top: Optional[int] = None,
):
    """
    Set window rectangle.

    This function sets the window size and position based on W3C WebDriver Specification.

    Args:
        conn: CDPConnectionSync: The connection to the websocket
        width: The desired window width in pixels
        height: The desired window height in pixels
        left: The desired window x coordinate
        top: The desired window y coordinate

    Returns:
        None

    Raises:
        WebDriverError: If the window rectangle setting fails
    """
    try:
        window_for_target = conn.execute(browser.get_window_for_target())
        window_id = window_for_target[0]
        bounds: Dict[str, Any] = {}
        if left is not None:
            bounds["left"] = left
        if top is not None:
            bounds["top"] = top
        bounds = browser.Bounds.from_json({"width": width, "height": height})
        conn.execute(browser.set_window_bounds(window_id=window_id, bounds=bounds))
    except Exception as e:
        raise WebDriverError("Failed to set window rectangle.") from e


def fullscreen_window(conn: SyncCDPConnection):
    """
    Fullscreen window.

    This function fullscreens the window based on W3C WebDriver Specification.

    Args:
        conn: CDPConnectionSync: The connection to the websocket

    Returns:
        None

    Raises:
        WebDriverError: If the fullscreen operation fails
    """
    try:
        _set_screen(conn, "normal")
        _set_screen(conn, "fullscreen")
    except Exception as e:
        raise WebDriverError("Failed to fullscreen window.") from e


def minimize_window(conn: SyncCDPConnection):
    """
    Minimize window.

    This function minimizes the window based on W3C WebDriver Specification.

    Args:
        conn: CDPConnectionSync: The connection to the websocket

    Returns:
        None

    Raises:
        WebDriverError: If the minimize operation fails
    """
    try:
        _set_screen(conn, "normal")
        _set_screen(conn, "minimized")
    except Exception as e:
        raise WebDriverError("Failed to minimize window.") from e


def maximize_window(conn: SyncCDPConnection):
    """
    Maximize window.

    This function maximizes the window based on W3C WebDriver Specification.

    Args:
        conn: CDPConnectionSync: The connection to the websocket

    Returns:
        None

    Raises:
        WebDriverError: If the maximize operation fails
    """
    try:
        _set_screen(conn, "normal")
        _set_screen(conn, "maximized")
    except Exception as e:
        raise WebDriverError("Failed to maximize window.") from e


def switch_to_window(conn: SyncCDPConnection, handle: target.TargetInfo):
    """
    Switch to window.

    This function switches the WebDriver context to a different window by its handle.
    Based on W3C WebDriver Specification.

    Args:
        conn: CDPConnectionSync: The connection to the websocket.

    Returns:
        new_conn (CDPConnectionSync): The new connection to the websocket.

    Raises:
        WebDriverError: If the switch to window operation fails.
    """
    if not handle:
        raise WebDriverError("Handle not informed")
    try:
        handles = get_window_handles(conn)
        index = 0
        for h in handles:
            if h.target_id == handle.target_id:
                break
            index += 1
        port = urlparse(conn.url).port
        with urllib.request.urlopen(f"http://127.0.0.1:{port}/json") as r:
            targets = json.loads(r.read())
            targets_page = [p for p in targets if p.get("type") == "page"]
            targets_page.reverse()
            ws_url = targets_page[index]["webSocketDebuggerUrl"]
            conn.set_url(ws_url)
            conn.connect()
            new_conn = conn
            _refresh_agents(new_conn)
            return new_conn
    except Exception as e:
        raise WebDriverError("Failed to switch to window.") from e


def new_window(conn: SyncCDPConnection) -> None:
    """
    Open a new window.

    This function opens a new window or tab based on W3C WebDriver Specification.

    Args:
        conn: CDPConnectionSync: The connection to the websocket
        window_type: The type of window to open ('tab' or 'window'). Defaults to 'tab'

    Returns:
        None

    Raises:
        WebDriverError: If the window creation fails
    """
    try:
        conn.execute(
            target.create_target(
                url="about:blank",
            )
        )
    except Exception as e:
        raise WebDriverError("Failed to open window.") from e


def switch_to_parent_frame(conn: SyncCDPConnection):
    """
    Switch to parent frame of 'element_frame'.

    This function switches the WebDriver context to the parent frame of the specified frame element.
    Based on W3C WebDriver Specification.

    Args:
        conn: CDPConnectionSync: The connection to the websocket
        element_frame: The frame element identifier whose parent frame to switch to

    Returns:
        None

    Raises:
        WebDriverError: If the switch to parent frame operation fails
    """
    try:
        doc = conn.execute(dom.get_document(depth=1))
        doc = GlobalValues.get_document_node()
        return doc.node_id
    except Exception as e:
        raise WebDriverError("Failed to switch to parent frame.") from e


def switch_to_frame(conn: SyncCDPConnection, element_frame: str) -> dom.NodeId:
    """
    Switch to frame 'element_frame'.

    This function switches the WebDriver context to the specified frame element.
    Based on W3C WebDriver Specification.

    Args:
        conn: CDPConnectionSync: The connection to the websocket
        element_frame: The frame element identifier to switch to

    Returns:
        cdp.dom.Node

    Raises:
        WebDriverError: If the switch to frame operation fails
    """
    try:
        desc = conn.execute(dom.describe_node(node_id=element_frame, depth=1))
        return desc.content_document.node_id
    except Exception as e:
        raise WebDriverError("Failed to switch to frame.") from e


def delete_all_cookies(conn: SyncCDPConnection):
    """
    Delete all cookies for the current session.

    This function removes all cookies associated with the active session,
    following the W3C WebDriver Specification for cookie management.

    Args:
        conn: CDPConnectionSync: The connection to the websocket.

    Returns:
        None

    Raises:
        WebDriverError: If the cookie deletion request fails or an error occurs
                        during the deletion process.
    """
    try:
        conn.execute(network.clear_browser_cookies())
    except Exception as e:
        raise WebDriverError("Failed to delete cookies.") from e


def get_alert_text(conn: SyncCDPConnection, element: dom.NodeId) -> str:
    """Get the text from an alert"""
    try:
        return _handle_alert(conn, element, timeout=TIMEOUT, function_callback=_accept_alert)
    except Exception as e:
        raise WebDriverError("Failed to get the alert text.") from e


def send_alert_text(
    conn: SyncCDPConnection, alert_element: dom.NodeId, text, timeout: float = TIMEOUT
):
    """
    Send text to an alert dialog.

    This function sends text to the currently open alert dialog and closes it.
    Based on W3C WebDriver Specification.

    Args:
        conn: CDPConnectionSync: The connection to the websocket
        alert_element: the alert element to send text
        text: The text to send to the alert dialog

    Returns:
        None

    Raises:
        WebDriverError: If sending text to the alert fails
    """
    try:
        _handle_alert(conn, alert_element, timeout, _send_test_to_prompt_alert, text)
    except Exception as e:
        raise WebDriverError("Failed to sent text to alert.") from e


def accept_alert(conn: SyncCDPConnection, alert_element: dom.NodeId, timeout: float = TIMEOUT):
    """
    Accept alert.

    This function accepts the currently open alert dialog.
    Based on W3C WebDriver Specification.

    Args:
        conn: CDPConnectionSync: The connection to the websocket
        alert_element: the alert to be accepted

    Returns:
        None

    Raises:
        WebDriverError: If the alert acceptance fails
    """
    try:
        _handle_alert(conn, alert_element, timeout, _accept_alert)
    except Exception as e:
        raise WebDriverError("Failed to accept alert.") from e


def dismiss_alert(conn: SyncCDPConnection, alert_element: dom.NodeId, timeout: float = TIMEOUT):
    """Dismiss alert

    This function dismisses the currently open alert dialog.
    Based on W3C WebDriver Specification.

    Args:
        conn: CDPConnectionSync: The connection to the websocket.
        alert_element: the alert to be dismissed

    Returns:
        None

    Raises:
        WebDriverError: If the alert dismissal fails.
    """
    try:
        _handle_alert(conn, alert_element, timeout, _dismiss_alert)
    except Exception as e:
        raise WebDriverError("Failed to dismiss alert.") from e


def take_screenshot_element(
    conn: SyncCDPConnection,
    element,
    path="/tmp",
    file_name="caqui",
):
    """Take screenshot of element

    Args:
        conn: CDPConnectionSync: The connection to the websocket..
        element: The identifier of the element to take a screenshot of.
        path: The directory path where the screenshot will be saved.
        file_name: The name of the file to save the screenshot as.

    Returns:
        None

    Raises:
        WebDriverError: If taking the screenshot fails.
    """
    try:
        box_model = conn.execute(dom.get_box_model(node_id=element))
        content_box = box_model.content[
            0:8
        ]  # Get the 8 points (x1, y1, x2, y2... etc) of the content box

        # Determine the clipping region coordinates from the box model points
        # The content_box provides points of a polygon, we need min/max x and y for the clip
        # Simplified assumption for a simple rectangle:
        x_coords = content_box[::2]
        y_coords = content_box[1::2]
        clip_x = min(x_coords)
        clip_y = min(y_coords)
        clip_width = max(x_coords) - clip_x
        clip_height = max(y_coords) - clip_y

        # Define the Clip object
        clip = page.Viewport(
            x=clip_x,
            y=clip_y,
            width=clip_width,
            height=clip_height,
            scale=1.0,  # Use 1.0 for a standard scale
        )

        now = datetime.datetime.now()
        timestamp = now.strftime(TIME_FORMAT_MICROSECONDS)

        # Capture the screenshot with the specified clip
        screenshot_data = conn.execute(page.capture_screenshot(clip=clip))
        save_picture(timestamp, path, file_name, screenshot_data)
    except Exception as e:
        raise WebDriverError("Failed to take screenshot from element.") from e


def take_screenshot(
    conn: SyncCDPConnection,
    path: str = "/tmp",
    file_name: str = "caqui",
):
    """Take screenshot

    Args:
        conn: CDPConnectionSync: The connection to the websocket..
        path: The directory path where the screenshot will be saved.
        file_name: The name of the file to save the screenshot as.

    Returns:
        NOne

    Raises:
        WebDriverError: If taking the screenshot fails.
    """
    try:
        # 2. Get the full layout metrics of the page
        # This command returns the content size of the page, regardless of the current viewport
        metrics = conn.execute(page.get_layout_metrics())
        content_size = metrics[-1]
        full_width = content_size.width
        full_height = content_size.height

        # 3. Override the device metrics to match the full page size
        # This effectively makes the entire page visible in a single "viewport"
        viewport = page.Viewport(0, 0, full_width, full_height, 1)
        conn.execute(
            emulation.set_device_metrics_override(
                width=full_width,
                height=full_height,
                device_scale_factor=1,
                mobile=False,
                screen_width=full_width,
                screen_height=full_height,
                position_x=0,
                position_y=0,
                viewport=viewport,
            )
        )

        screenshot_data = conn.execute(page.capture_screenshot(quality=90))

        # 5. Reset viewport emulation (optional, good practice)
        conn.execute(emulation.clear_device_metrics_override())

        # 6. Save the screenshot data to a file
        now = datetime.datetime.now()
        timestamp = now.strftime(TIME_FORMAT_MICROSECONDS)
        save_picture(timestamp, path, file_name, screenshot_data)
    except Exception as e:
        raise WebDriverError("Failed to take screenshot.") from e


def get_named_cookie(conn, name) -> dict:
    """Get cookie by name.

    This function retrieves a cookie from the WebDriver session based on the specified name.

    Args:
        conn: CDPConnectionSync: The connection to the websocket..
        name: The name of the cookie to retrieve.

    Returns:
        A dictionary representing the cookie if found, otherwise an empty dictionary.

    Raises:
        WebDriverError: If the request to get the cookie fails.
    """
    try:
        cookies = get_cookies(conn)
        return [c for c in cookies if c.get("name") == name][0]
    except Exception as e:
        raise WebDriverError(f"Failed to get cookie '{name}'.") from e


def get_computed_label(conn: SyncCDPConnection, element: dom.NodeId) -> str:
    """Get the element tag computed label. Get the accessibility name.

    Args:
        conn: CDPConnectionSync: The connection to the websocket..
        element: The identifier of the element to retrieve the computed label for.

    Returns:
        The computed label of the element, which represents its accessibility name.

    Raises:
        WebDriverError: If retrieving the computed label fails.
    """
    try:
        # Query the accessibility tree for the specific node
        # The queryAXTree command takes a nodeId (among other possible identifiers)
        ax_nodes_response = conn.execute(accessibility.get_partial_ax_tree(node_id=element))

        # The response contains a list of AXNode objects
        # The target node's info should be in the list
        for ax_node in ax_nodes_response:
            # Check if this is the correct node and extract its name/label
            # The 'name' property holds the computed accessible name
            if ax_node.name.value:
                return ax_node.name.value
        return ""
    except Exception as e:
        raise WebDriverError("Failed to get element computed label.") from e


def get_computed_role(conn: SyncCDPConnection, element: dom.NodeId) -> str:
    """Get the element tag computed role (the element role).

    Args:
        conn: CDPConnectionSync: The connection to the websocket..
        element: The identifier of the element to retrieve the computed role for.

    Returns:
        A string representing the computed role

    Raises:
        WebDriverError: If retrieving the computed role fails.
    """
    try:
        # Get accessibility information for the node, including the role
        # We use getAXNodeAndAncestors to retrieve the AXNode object which holds computed properties
        ax_nodes_response = conn.execute(accessibility.get_ax_node_and_ancestors(node_id=element))

        # The response is a list of nodes, the first one is the target node
        if ax_nodes_response:
            target_ax_node = ax_nodes_response[0]
            computed_role = target_ax_node.role.value
            return computed_role
        return ""
    except Exception as e:
        raise WebDriverError("Failed to get element computed label.") from e


def get_tag_name(conn: SyncCDPConnection, element: dom.NodeId) -> str:
    """
    Get the tag name of a specified element in a WebDriver session.

    Parameters:
        conn: CDPConnectionSync: The connection to the websocket.
        element: The identifier for the specific element whose tag name is to be retrieved.

    Returns:
        A string representing the tag name of the specified element.

    Raises:
        WebDriverError: If there is an error while attempting to retrieve the element's tag name.
    """
    try:
        result = conn.execute(dom.describe_node(node_id=element))
        return result.node_name
    except Exception as e:
        raise WebDriverError("Failed to get element name.") from e


# def get_shadow_element_v1(
#     conn: CDPConnectionSync,
#     shadow_root_locator_type: str,
#     shadoe_root_locator_value: str,
#     locator_type: str,
#     locator_value: str,
# ) -> dom.NodeId:
#     """
#     Get the shadow root element from a web page using the W3C WebDriver Specification.

#     Parameters:
#         conn: CDPConnectionSync: The connection to the websocket.
#         shadow_element: The ID or name of the shadow element to retrieve.
#         locator_type: The type of locator to use (e.g., 'css selector', 'xpath').
#         locator_value: The value of the locator to find the element.

#     Returns:
#         The shadow root element as a string, or an empty string if not found.

#     Raises:
#         WebDriverError: If there is an error retrieving the shadow element.
#     """
#     try:
#         host_node = _find_element(
#             conn, shadow_root_locator_type, shadoe_root_locator_value, depth=-1, pierce=True
#         )
#         host_details = conn.execute(dom.describe_node(node_id=host_node, depth=1))
#         shadow_roots = host_details.shadow_roots
#         return _find_element(conn, locator_type, locator_value, shadow_roots[0].node_id)
#     except Exception as e:
#         raise WebDriverError("Failed to get the element shadow.") from e


def get_shadow_element(
    conn: SyncCDPConnection,
    shadow_element,
    locator_type: str,
    locator_value: str,
) -> dom.NodeId:
    """
    Get the shadow root element from a web page using the W3C WebDriver Specification.

    Parameters:
        conn: CDPConnectionSync: The connection to the websocket.
        shadow_element: The ID or name of the shadow element to retrieve.
        locator_type: The type of locator to use (e.g., 'css selector', 'xpath').
        locator_value: The value of the locator to find the element.

    Returns:
        The shadow root element as a string, or an empty string if not found.

    Raises:
        WebDriverError: If there is an error retrieving the shadow element.
    """
    try:
        host_details = conn.execute(dom.describe_node(node_id=shadow_element, depth=1))
        shadow_roots = host_details.shadow_roots
        return _find_element(conn, locator_type, locator_value, shadow_roots[0].node_id)
    except Exception as e:
        raise WebDriverError("Failed to get the element shadow.") from e


def get_shadow_elements(
    conn: SyncCDPConnection,
    shadow_element: str,
    locator_type: str,
    locator_value: str,
) -> List[str]:
    """
    Get the list of shadow root elements.

    Args:
        conn: CDPConnectionSync: The connection to the websocket.
        shadow_element: The identifier for the shadow element to retrieve.
        locator_type: The type of locator to use (e.g., 'css selector', 'xpath').
        locator_value: The value of the locator to find the element.

    Returns:
        A list of shadow root element identifiers.

    Raises:
        WebDriverError: If there is an error retrieving the shadow elements.
    """
    """Get the list of shadow root elements"""
    try:
        host_details = conn.execute(dom.describe_node(node_id=shadow_element, depth=1))
        shadow_roots = host_details.shadow_roots
        return _find_all_elements(conn, locator_type, locator_value, shadow_roots[0].node_id)
    except Exception as e:
        raise WebDriverError("Failed to get the element shadow.") from e


def get_rect(conn: SyncCDPConnection, element) -> dict:
    """Get the element rectangle"""
    try:
        result = conn.execute(dom.get_box_model(node_id=element))
        # Quad([205, 103.4375, 226.484375, 103.4375, 226.484375, 118.4375, 205, 118.4375])
        content = result.content
        height = abs(content[-1] - content[1])
        width = abs(content[2] - content[0])
        x = content[0]
        y = content[1]
        return {"height": height, "width": width, "x": x, "y": y}
    except Exception as e:
        raise WebDriverError("Failed to get element rect.") from e


def actions_move_to_element(conn: SyncCDPConnection, element: str) -> None:
    """
    Move to an element simulating a mouse movement.

    This function sends a WebDriver Actions command to move the mouse pointer to a
    specified element, following the W3C WebDriver Specification.

    Args:
        conn: CDPConnectionSync: The connection to the websocket..
        element: The element identifier (W3C element reference) to move the pointer to.

    Returns:
        None

    Raises:
        WebDriverError: If the action fails to move to the element.

    Note:
        The mouse movement is performed with zero duration, resulting in an instant
        pointer movement to the element's location. The element reference follows the
        W3C WebDriver specification format.
    """
    """Move to an element simulating a mouse movement"""
    try:
        x, y = _get_element_center(conn, element)
        conn.execute(
            input_.dispatch_mouse_event(
                "mouseMoved",
                x=x,
                y=y,
            )
        )
    except Exception as e:
        raise WebDriverError("Failed to move to element.") from e


def actions_scroll_to_element(
    conn: SyncCDPConnection,
    element,
):
    """Scroll to an element simulating a mouse movement"""
    try:
        remote_object = conn.execute(dom.resolve_node(node_id=element))
        conn.execute(
            runtime.call_function_on(
                function_declaration="""
                    function() {
                        this.scrollIntoView({
                            block: 'center',
                            inline: 'center',
                            behavior: 'instant'
                        });
                    }""",
                object_id=remote_object.object_id,
            )
        )
    except Exception as e:
        raise WebDriverError("Failed to scroll to element.") from e


def submit(conn: SyncCDPConnection, element):
    """Submit a form. It is similar to 'submit' funtion in Seleniu
    It is not part of W3C WebDriver. Just added for convenience
    """
    try:
        remote_object = conn.execute(dom.resolve_node(node_id=element))
        conn.execute(
            runtime.call_function_on(
                function_declaration="""
                    function() {
                        this.submit();
                    }""",
                object_id=remote_object.object_id,
            )
        )
    except Exception as e:
        raise WebDriverError("Failed to submit form.") from e


def find_children_elements(
    conn: SyncCDPConnection,
    parent_element: str,
    locator_type: str,
    locator_value: str,
):
    """Find the children elements by 'locator_type'

    If the 'parent_element' is a shadow element, set the 'locator_type' as 'id' or
    'css selector'
    """
    try:
        return _find_all_elements(conn, locator_type, locator_value, parent_element)
    except Exception as e:
        raise WebDriverError(
            f"Failed to find the children elements from '{parent_element}'."
        ) from e


def find_child_element(
    conn: SyncCDPConnection,
    parent_element: str,
    locator_type: str,
    locator_value: str,
):
    """Find the child element by 'locator_type'"""
    try:
        return find_element(conn, locator_type, locator_value, parent_element)
    except Exception as e:
        raise WebDriverError(f"Failed to find the child element from '{parent_element}'.") from e


def get_page_source(conn: SyncCDPConnection) -> str:
    """Get the page source (all content)"""
    try:
        doc = GlobalValues.get_document_node()
        return conn.execute(dom.get_outer_html(node_id=doc.node_id))
    except Exception as e:
        raise WebDriverError("Failed to get the page source.") from e


def execute_script(
    conn: SyncCDPConnection,
    element: dom.NodeId,
    script: str,
    positive=True,
    return_by_value=False,
):
    """Executes a script, like 'style.background='#000000'' to change
    the background of the element"""
    positive = "" if positive else "!"
    try:
        remote_object = conn.execute(dom.resolve_node(node_id=element))
        return conn.execute(
            runtime.call_function_on(
                function_declaration="""
                    function() {"""
                f"""
                        return {positive}this.{script};
                    """
                """
                    }
                """,
                object_id=remote_object.object_id,
                return_by_value=return_by_value,
            )
        )
    except Exception as e:
        raise WebDriverError("Failed to execute script.") from e


def get_active_element(conn: SyncCDPConnection):
    """Get the active element"""
    try:
        eval_result = conn.execute(runtime.evaluate(expression="document.activeElement"))
        object_id = eval_result[0].object_id
        if not object_id:
            raise WebDriverError("No active element or it is the document body.")
        node_id_response = conn.execute(dom.request_node(object_id=object_id))
        return node_id_response
    except Exception as e:
        raise WebDriverError("Failed to check if element is selected.") from e


def clear_element(conn: SyncCDPConnection, element: dom.NodeId):
    """Clear the element text"""
    try:
        resolved = conn.execute(dom.resolve_node(node_id=element))

        conn.execute(
            runtime.call_function_on(
                object_id=resolved.object_id,
                function_declaration="""
                    function() {
                        this.value = '';
                        this.dispatchEvent(new Event('input', { bubbles: true }));
                        this.dispatchEvent(new Event('change', { bubbles: true }));
                    }
                """,
            )
        )
    except Exception as e:
        raise WebDriverError("Failed to clear the element text.") from e


def is_element_enabled(conn: SyncCDPConnection, element) -> bool:
    """Check if element is enabled"""
    try:
        script = "disabled"
        result = execute_script(conn, element, script, positive=False, return_by_value=True)
        return bool(result[0].value)
    except Exception as e:
        raise WebDriverError("Failed to check if element is enabled.") from e


def get_window_rectangle(conn: SyncCDPConnection) -> dict:
    """Get window rectangle"""
    try:
        _, bounds = conn.execute(browser.get_window_for_target())
        return {
            "x": bounds.left,
            "y": bounds.top,
            "width": bounds.width,
            "height": bounds.height,
        }
    except Exception as e:
        raise WebDriverError("Failed to get window rectangle.") from e


def get_window_handles(conn: SyncCDPConnection) -> List[target.TargetInfo]:
    """Get window handles"""
    try:
        targets = conn.execute(target.get_targets())
        return [t for t in targets if t.type_ == "page"]
    except Exception as e:
        raise WebDriverError("Failed to get window handles.") from e


def close_window(conn: SyncCDPConnection):
    """Close active window"""
    try:
        new_window(conn)
        conn.execute(target.page.close())
    except Exception as e:
        raise WebDriverError("Failed to close active window.") from e


def get_window(conn) -> str:
    """Get window"""
    try:
        result = get_window_handles(conn)
        return result[0].target_id
    except Exception as e:
        raise WebDriverError("Failed to get window.") from e


def get_status(conn: SyncCDPConnection) -> dict:
    """Returns the status and details of the WebDriver"""
    try:
        conn.execute(target.get_targets())
        return {"ready": True}
    except Exception as e:
        raise WebDriverError("Failed to get status.") from e



import caqui._vendor.chrome_devtools_protocol.cdp.accessibility
import caqui._vendor.chrome_devtools_protocol.cdp.animation
import caqui._vendor.chrome_devtools_protocol.cdp.audits
import caqui._vendor.chrome_devtools_protocol.cdp.autofill
import caqui._vendor.chrome_devtools_protocol.cdp.background_service
import caqui._vendor.chrome_devtools_protocol.cdp.bluetooth_emulation
import caqui._vendor.chrome_devtools_protocol.cdp.browser
import caqui._vendor.chrome_devtools_protocol.cdp.cache_storage
import caqui._vendor.chrome_devtools_protocol.cdp.cast
import caqui._vendor.chrome_devtools_protocol.cdp.connection
import caqui._vendor.chrome_devtools_protocol.cdp.console
import caqui._vendor.chrome_devtools_protocol.cdp.css
import caqui._vendor.chrome_devtools_protocol.cdp.debugger
import caqui._vendor.chrome_devtools_protocol.cdp.device_access
import caqui._vendor.chrome_devtools_protocol.cdp.device_orientation
import caqui._vendor.chrome_devtools_protocol.cdp.dom
import caqui._vendor.chrome_devtools_protocol.cdp.dom_debugger
import caqui._vendor.chrome_devtools_protocol.cdp.dom_snapshot
import caqui._vendor.chrome_devtools_protocol.cdp.dom_storage
import caqui._vendor.chrome_devtools_protocol.cdp.emulation
import caqui._vendor.chrome_devtools_protocol.cdp.event_breakpoints
import caqui._vendor.chrome_devtools_protocol.cdp.extensions
import caqui._vendor.chrome_devtools_protocol.cdp.fed_cm
import caqui._vendor.chrome_devtools_protocol.cdp.fetch
import caqui._vendor.chrome_devtools_protocol.cdp.file_system
import caqui._vendor.chrome_devtools_protocol.cdp.headless_experimental
import caqui._vendor.chrome_devtools_protocol.cdp.heap_profiler
import caqui._vendor.chrome_devtools_protocol.cdp.indexed_db
import caqui._vendor.chrome_devtools_protocol.cdp.input_
import caqui._vendor.chrome_devtools_protocol.cdp.inspector
import caqui._vendor.chrome_devtools_protocol.cdp.io
import caqui._vendor.chrome_devtools_protocol.cdp.layer_tree
import caqui._vendor.chrome_devtools_protocol.cdp.log
import caqui._vendor.chrome_devtools_protocol.cdp.media
import caqui._vendor.chrome_devtools_protocol.cdp.memory
import caqui._vendor.chrome_devtools_protocol.cdp.network
import caqui._vendor.chrome_devtools_protocol.cdp.overlay
import caqui._vendor.chrome_devtools_protocol.cdp.page
import caqui._vendor.chrome_devtools_protocol.cdp.performance
import caqui._vendor.chrome_devtools_protocol.cdp.performance_timeline
import caqui._vendor.chrome_devtools_protocol.cdp.preload
import caqui._vendor.chrome_devtools_protocol.cdp.profiler
import caqui._vendor.chrome_devtools_protocol.cdp.pwa
import caqui._vendor.chrome_devtools_protocol.cdp.runtime
import caqui._vendor.chrome_devtools_protocol.cdp.schema
import caqui._vendor.chrome_devtools_protocol.cdp.security
import caqui._vendor.chrome_devtools_protocol.cdp.service_worker
import caqui._vendor.chrome_devtools_protocol.cdp.storage
import caqui._vendor.chrome_devtools_protocol.cdp.system_info
import caqui._vendor.chrome_devtools_protocol.cdp.target
import caqui._vendor.chrome_devtools_protocol.cdp.tethering
import caqui._vendor.chrome_devtools_protocol.cdp.tracing
import caqui._vendor.chrome_devtools_protocol.cdp.util
import caqui._vendor.chrome_devtools_protocol.cdp.web_audio
import caqui._vendor.chrome_devtools_protocol.cdp.web_authn



# Copyright (C) 2023 Caqui - All Rights Reserved
# You may use, distribute and modify this code under the
# terms of the MIT license.
# Visit: https://github.com/douglasdcm/caqui

from typing import TYPE_CHECKING, Coroutine, List, Union

from caqui.webdriver.engine import asynchronous

if TYPE_CHECKING:
    from caqui.webdriver.drivers import AsyncDriver
    from caqui.webdriver.element import Element


class ActionChainsW3C:
    def __init__(self, driver: "AsyncDriver") -> None:
        self._server_url = driver.server_url
        self._session = driver.session
        self._session_http = driver.session_http
        self._coroutines: List[Coroutine] = []
        self._element = Union["Element", None]

    def click(self, element: "Element") -> "ActionChainsW3C":
        """
        Clicks on the element `element`
        """
        self._element = element
        coroutine = asynchronous.click(
            self._server_url, self._session, str(element), session_http=self._session_http
        )
        self._coroutines.append(coroutine)
        return self

    def move_to_element(self, element: "Element") -> "ActionChainsW3C":
        """Move the mouse to the element `element`"""
        self._element = element
        coroutine = asynchronous.actions_move_to_element(
            self._server_url, self._session, str(element), session_http=self._session_http
        )
        self._coroutines.append(coroutine)
        return self

    def scroll_to_element(self, element: "Element", delta_y: int = 1000) -> "ActionChainsW3C":
        """Scrolls the screen to the element `element`"""
        self._element = element
        coroutine = asynchronous.actions_scroll_to_element(
            self._server_url,
            self._session,
            str(element),
            delta_y=delta_y,
            session_http=self._session_http,
        )
        self._coroutines.append(coroutine)
        return self

    async def perform(self) -> None:
        """Executes the chain of Coroutines"""
        [await coroutine for coroutine in self._coroutines]


class ActionChainsJsonWire(ActionChainsW3C):
    def __init__(self, driver):
        super().__init__(driver)

    def move_to_element(self, element: "Element") -> "ActionChainsW3C":
        """Move the mouse to the element `element`"""
        self._element = element
        coroutine = asynchronous.actions_move_to_element_jsonwire(
            self._server_url, self._session, str(element), session_http=self._session_http
        )
        self._coroutines.append(coroutine)
        return self

    def scroll_to_element(self, element: "Element", delta_y=1000) -> "ActionChainsW3C":
        """Scrolls the screen to the element `element`"""
        self._element = element
        coroutine = asynchronous.actions_scroll_to_element_jsonwire(
            self._server_url,
            self._session,
            str(element),
            delta_y=delta_y,
            session_http=self._session_http,
        )
        self._coroutines.append(coroutine)
        return self



# Copyright (C) 2023 Caqui - All Rights Reserved
# You may use, distribute and modify this code under the
# terms of the MIT license.
# Visit: https://github.com/douglasdcm/caqui

from typing import TYPE_CHECKING, Coroutine, List, Union

from caqui.cdp.engine import asynchronous

if TYPE_CHECKING:
    from caqui.cdp.asynchronous.drivers import AsyncDriverCDP
    from caqui.cdp.asynchronous.element import Element


class ActionChains:
    def __init__(self, driver: "AsyncDriverCDP") -> None:
        self._conn = driver.conn
        self._coroutines: List[Coroutine] = []
        self._element = Union["Element", None]

    def click(self, element: "Element") -> "ActionChains":
        """
        Clicks on the element `element`
        """
        self._element = element
        coroutine = asynchronous.click(
            self._conn,
            element.element_id,
        )
        self._coroutines.append(coroutine)
        return self

    def move_to_element(self, element: "Element") -> "ActionChains":
        """Move the mouse to the element `element`"""
        self._element = element
        coroutine = asynchronous.actions_move_to_element(
            self._conn,
            element.element_id,
        )
        self._coroutines.append(coroutine)
        return self

    def scroll_to_element(self, element: "Element", delta_y: int = 1000) -> "ActionChains":
        """Scrolls the screen to the element `element`"""
        self._element = element
        coroutine = asynchronous.actions_scroll_to_element(
            self._conn,
            element.element_id,
        )
        self._coroutines.append(coroutine)
        return self

    async def perform(self) -> None:
        """Executes the chain of Coroutines"""
        [await coroutine for coroutine in self._coroutines]



# Copyright (C) 2025 python-chrome-devtools-protocol - All Rights Reserved
# You may use, distribute and modify this code under the
# terms of the MIT license.
# Visit: https://github.com/HyperionGray/python-chrome-devtools-protocol/pull/59/files
#
# Inportant note!
# This code was copied from the cdp library and changed to allow multiple connections


"""
CDP Connection Module

This module provides I/O and multiplexing support for Chrome DevTools Protocol.
It handles WebSocket connections, JSON-RPC message framing, command multiplexing,
and event dispatching.
"""

from __future__ import annotations

import asyncio
import json
import logging
import queue
import threading
import time
import typing
import uuid
from dataclasses import dataclass

import websocket
from websocket import create_connection

try:
    import websockets
    from websockets.client import WebSocketClientProtocol

    WEBSOCKETS_AVAILABLE = True
except ImportError:
    WEBSOCKETS_AVAILABLE = False
    WebSocketClientProtocol = typing.Any  # type: ignore

from caqui._vendor.chrome_devtools_protocol.cdp.util import T_JSON_DICT, parse_json_event

logger = logging.getLogger(__name__)


class CDPError(Exception):
    """Base exception for CDP errors."""

    pass


class CDPConnectionError(CDPError):
    """Raised when there's a connection error."""

    pass


class CDPCommandError(CDPError):
    """Raised when a command returns an error."""

    def __init__(self, code: int, message: str, data: typing.Optional[typing.Any] = None):
        self.code = code
        self.message = message
        self.data = data
        super().__init__(f"CDP Command Error {code}: {message}")


@dataclass
class PendingCommand:
    """Represents a command waiting for a response."""

    future: asyncio.Future
    method: str
    params: T_JSON_DICT


class AsyncCDPConnection:
    """
    Manages a WebSocket connection to Chrome DevTools Protocol.

    This class handles:
    - WebSocket connection management
    - JSON-RPC message framing (request ID assignment)
    - Command multiplexing (tracking multiple concurrent commands)
    - Event dispatching
    - Error handling

    Example:
        async with CDPConnection("ws://localhost:9222/devtools/page/...") as conn:
            # Send a command
            result = await conn.execute(some_command())

            # Listen for events
            async for event in conn.listen():
                print(event)
    """

    main_context_id = None

    def __init__(self, url: str, timeout: float = 30.0):
        """
        Initialize a CDP connection.

        Args:
            url: WebSocket URL for the CDP endpoint
            timeout: Default timeout for commands in seconds
        """
        if not WEBSOCKETS_AVAILABLE:
            raise ImportError(
                "websockets library is required for CDPConnection. "
                "Install it with: pip install websockets"
            )

        self.url = url
        self.timeout = timeout
        self._ws: typing.Optional[WebSocketClientProtocol] = None
        self._next_command_id = 1
        self._pending_commands: typing.Dict[int, PendingCommand] = {}
        self._event_queue: asyncio.Queue = asyncio.Queue()
        self._recv_task: typing.Optional[asyncio.Task] = None
        self._closed = False
        self._wss: typing.List[WebSocketClientProtocol] = []
        self._loop = None

    async def set_url(self, url):
        self.url = url
        self._ws = None

    async def connect(self) -> None:
        """Establish the WebSocket connection."""
        if self._ws is not None:
            raise CDPConnectionError("Already connected")

        try:
            self._ws = await websockets.connect(self.url)  # type: ignore
            self._wss.append(self._ws)
            self._recv_task = asyncio.create_task(self._receive_loop())
            logger.info(f"Connected to {self.url}")
        except Exception as e:
            raise CDPConnectionError(f"Failed to connect to {self.url}: {e}")

    async def close(self) -> None:
        """Close the WebSocket connection."""
        if self._closed:
            return

        self._closed = True

        # Cancel the receive task
        if self._recv_task:
            self._recv_task.cancel()
            try:
                await self._recv_task
            except asyncio.CancelledError:
                pass

        # Cancel all pending commands
        for cmd_id, pending in self._pending_commands.items():
            if not pending.future.done():
                pending.future.cancel()
        self._pending_commands.clear()

        # Close the WebSocket
        for ws in self._wss:
            await ws.close()

        if self._ws:
            self._ws = None

        logger.info("Connection closed")

    async def __aenter__(self) -> AsyncCDPConnection:
        """Async context manager entry."""
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """Async context manager exit."""
        await self.close()

    async def _receive_loop(self) -> None:
        """
        Main receive loop that processes incoming WebSocket messages.

        This loop:
        - Receives messages from the WebSocket
        - Parses JSON-RPC responses and matches them to pending commands
        - Dispatches events to the event queue
        """
        try:
            while not self._closed and self._ws:
                try:
                    message = await self._ws.recv()
                    data = json.loads(message)

                    if "id" in data:
                        # This is a command response
                        await self._handle_response(data)
                    elif "method" in data:
                        # This is an event
                        await self._handle_event(data)
                    else:
                        logger.warning(f"Received unexpected message: {data}")

                except json.JSONDecodeError as e:
                    logger.error(f"Failed to decode JSON: {e}")
                except Exception as e:
                    logger.error(f"Error in receive loop: {e}")
                    if not self._closed:
                        raise
        except asyncio.CancelledError:
            logger.debug("Receive loop cancelled")
        except Exception as e:
            logger.error(f"Fatal error in receive loop: {e}")
            # Cancel all pending commands with this error
            for pending in self._pending_commands.values():
                if not pending.future.done():
                    pending.future.set_exception(CDPConnectionError(f"Connection error: {e}"))

    async def _handle_response(self, data: T_JSON_DICT) -> None:
        """Handle a command response."""
        cmd_id = data["id"]

        if cmd_id not in self._pending_commands:
            logger.warning(f"Received response for unknown command ID {cmd_id}")
            return

        pending = self._pending_commands.pop(cmd_id)

        if "error" in data:
            error = data["error"]
            exc = CDPCommandError(
                code=error.get("code", -1),
                message=error.get("message", "Unknown error"),
                data=error.get("data"),
            )
            pending.future.set_exception(exc)
        else:
            result = data.get("result", {})
            pending.future.set_result(result)

    async def _handle_event(self, data: T_JSON_DICT) -> None:
        """Handle an event notification."""
        try:
            event = parse_json_event(data)
            await self._event_queue.put(event)
        except Exception as e:
            logger.error(f"Failed to parse event: {e}")

    async def execute(
        self,
        cmd: typing.Generator[T_JSON_DICT, T_JSON_DICT, typing.Any],
        timeout: typing.Optional[float] = None,
    ) -> typing.Any:
        """
        Execute a CDP command.

        This method:
        - Assigns a unique ID to the command
        - Sends it over the WebSocket
        - Waits for the response (with multiplexing support)
        - Returns the parsed result

        Args:
            cmd: A CDP command generator (from any CDP domain module)
            timeout: Optional timeout override for this command

        Returns:
            The command result (type depends on the command)

        Raises:
            CDPCommandError: If the command returns an error
            asyncio.TimeoutError: If the command times out
            CDPConnectionError: If there's a connection error

        Example:
            from caqui._vendor.chrome_devtools_protocol.cdp import page
            result = await conn.execute(page.navigate(url="https://example.com"))
        """
        if self._ws is None:
            raise CDPConnectionError("Not connected")

        if self._closed:
            raise CDPConnectionError("Connection closed")

        # Get the command request from the generator
        request = cmd.send(None)  # type: ignore[arg-type]

        # Assign a unique ID
        cmd_id = self._next_command_id
        self._next_command_id += 1
        request["id"] = cmd_id

        # Create a future to track this command
        future: asyncio.Future = asyncio.Future()
        self._pending_commands[cmd_id] = PendingCommand(
            future=future, method=request["method"], params=request.get("params", {})
        )

        try:
            # Send the command
            await self._ws.send(json.dumps(request))
            logger.debug(f"Sent command {cmd_id}: {request['method']}")
            # Wait for the response
            timeout_val = timeout if timeout is not None else self.timeout
            result = await asyncio.wait_for(future, timeout=timeout_val)

            # Send the result back to the generator
            try:
                # print(request)
                # print(result)
                # print("=================")
                cmd.send(result)
            except StopIteration as e:
                return e.value

            raise CDPError("Command generator did not stop")

        except asyncio.TimeoutError:
            # Clean up the pending command
            self._pending_commands.pop(cmd_id, None)
            raise asyncio.TimeoutError(f"Command {request['method']} timed out")
        except Exception:
            # Clean up the pending command on error
            self._pending_commands.pop(cmd_id, None)
            raise

    async def listen(self) -> typing.AsyncIterator[typing.Any]:
        """
        Listen for events from the browser.

        This is an async iterator that yields CDP events as they arrive.

        Yields:
            CDP event objects (type depends on the event)

        Example:
            async for event in conn.listen():
                if isinstance(event, page.LoadEventFired):
                    print("Page loaded!")
        """
        while not self._closed:
            try:
                event = await asyncio.wait_for(self._event_queue.get(), timeout=1.0)
                yield event
            except asyncio.TimeoutError:
                # Check if connection is still alive
                if self._closed:
                    break
                continue

    def get_event_nowait(self) -> typing.Optional[typing.Any]:
        """
        Get an event from the queue without waiting.

        Returns:
            A CDP event object, or None if no events are available
        """
        try:
            return self._event_queue.get_nowait()
        except asyncio.QueueEmpty:
            return None

    @property
    def is_connected(self) -> bool:
        """Check if the connection is open."""
        return self._ws is not None and not self._closed

    @property
    def pending_command_count(self) -> int:
        """Get the number of pending commands (for debugging/monitoring)."""
        return len(self._pending_commands)


class SyncCDPConnection:
    def __init__(self, url: str, timeout: float = 30.0):
        self._url = url
        self._timeout = timeout
        self._ws = None
        self._request_id = 0
        self._message_queue = queue.Queue()

    def __enter__(self) -> SyncCDPConnection:
        """Sync context manager entry."""
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Sync context manager exit."""
        self.close()

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        self._url = value

    def set_url(self, url):
        self._url = url

    def connect(self):
        self._ws = create_connection(self._url)
        t = threading.Thread(target=run_receive_thread, args=(self._ws,))
        t.daemon = True
        t.start()

    def get_event_nowait(self) -> typing.Optional[typing.Any]:
        """
        Get an event from the queue without waiting.

        Returns:
            A CDP event object, or None if no events are available
        """
        try:
            return _event_queue.get_nowait()
        except queue.Empty:
            return None

    def execute(self, command, wait=True):
        try:
            result = None
            request = next(command)
            self._request_id += 1

            request["id"] = self._request_id
            self._ws.send(json.dumps(request))
            if not wait:
                return
            timeout = 10  # seconds
            start_time = time.time()
            request_id = self._request_id
            pending_requests[request_id] = None
            while pending_requests[request_id] is None:
                if time.time() - start_time > timeout:
                    del pending_requests[request_id]
                    raise TimeoutError("Response not received within timeout")
                # Process messages from the queue in the main thread while waiting
                while not message_queue.empty():
                    process_incoming_message()
                # time.sleep(0.1) # Prevents busy waiting

            response = pending_requests[request_id]
            raw = response
            del pending_requests[request_id]
            # return response
            # raw = json.loads(self._ws.recv())
            if "error" in raw:
                raise CDPCommandError(raw["error"], message=raw)
            try:
                # print(request)
                # print(raw)
                # print("------------------")
                result = raw.get("result", {})
                if result:
                    # time.sleep(0.1)
                    command.send(result)
                    raise RuntimeError("CDP generator did not stop")
            except StopIteration as e:
                return e.value
        except Exception:
            raise

    def close(self):
        self._ws.close()


# A dictionary to store pending requests and their expected responses
pending_requests = {}
# A queue to put incoming messages for processing
message_queue = queue.Queue()
_event_queue = queue.Queue()


def on_message(ws, message):
    message_queue.put(message)


def _handle_event(data: T_JSON_DICT) -> None:
    """Handle an event notification."""
    try:
        event = parse_json_event(data)
        _event_queue.put(event)
    except Exception as e:
        logger.error(f"Failed to parse event: {e}")


def run_receive_thread(ws):
    # This loop runs indefinitely in a separate thread to receive messages
    while True:
        try:
            message = ws.recv()
            data = json.loads(message)
            if "method" in data:
                # This is an event
                _handle_event(data)
            if message:
                on_message(ws, message)
        except websocket.WebSocketConnectionClosedException:
            break
        except Exception as e:
            logging.exception(f"Error in receive thread: {e}")
            break


def send_request_and_wait_for_response(ws, request_data):
    request_id = str(uuid.uuid4())  # Generate a unique ID
    request_data["id"] = request_id

    # Store a placeholder for the response
    pending_requests[request_id] = None

    ws.send(json.dumps(request_data))

    # Wait for the response in a blocking manner
    timeout = 10  # seconds
    start_time = time.time()
    while pending_requests[request_id] is None:
        if time.time() - start_time > timeout:
            del pending_requests[request_id]
            raise TimeoutError("Response not received within timeout")
        # Process messages from the queue in the main thread while waiting
        while not message_queue.empty():
            process_incoming_message()
        # time.sleep(0.1) # Prevents busy waiting

    response = pending_requests[request_id]
    del pending_requests[request_id]
    return response


def process_incoming_message():
    # Called by the main thread to handle messages received by the receive thread
    message_str = message_queue.get()
    try:
        message = json.loads(message_str)
        msg_id = message.get("id")
        if msg_id in pending_requests:
            # If it matches a pending request, store it
            pending_requests[msg_id] = message
    except json.JSONDecodeError:
        logging.exception(f"Could not decode JSON: {message_str}")
